from types import TracebackType
from typing import Optional, Self, Type

from httpx import AsyncBaseTransport, Request, Response
from whistle import IAsyncEventDispatcher

from harp import get_logger
from harp_apps.http_client.events import (
    EVENT_FILTER_HTTP_CLIENT_REQUEST,
    EVENT_FILTER_HTTP_CLIENT_RESPONSE,
    HttpClientFilterEvent,
)

logger = get_logger(__name__)


class AsyncFilterableTransport(AsyncBaseTransport):
    def __init__(self, transport: AsyncBaseTransport, dispatcher: IAsyncEventDispatcher):
        self._transport = transport
        self._dispatcher = dispatcher

    async def handle_async_request(self, request: Request) -> Response:
        event = HttpClientFilterEvent(request)
        await self._dispatcher.adispatch(EVENT_FILTER_HTTP_CLIENT_REQUEST, event)
        logger.debug(f"▶▶▶ {event.request}", headers=event.request.headers)
        if not event.response:
            event.response = await self._transport.handle_async_request(event.request)
        await self._dispatcher.adispatch(EVENT_FILTER_HTTP_CLIENT_RESPONSE, event)
        logger.debug(
            f"◀◀◀ {event.response}",
            cache_control=event.response.headers.get("cache-control"),
        )
        return event.response

    async def aclose(self) -> None:
        await self._transport.aclose()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_value: Optional[BaseException] = None,
        traceback: Optional[TracebackType] = None,
    ) -> None:
        await self.aclose()



from datetime import UTC, datetime
from typing import TYPE_CHECKING, Optional

from sqlalchemy import TIMESTAMP, ForeignKey, Index, Integer, String, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship

from harp.http import BaseMessage, get_serializer_for
from harp.models.messages import Message as MessageModel

from .base import Base, Repository, with_session

if TYPE_CHECKING:
    from .transactions import Transaction


class Message(Base):
    __tablename__ = "messages"

    id: Mapped[int] = mapped_column(Integer(), primary_key=True, unique=True, autoincrement=True)
    kind: Mapped[str] = mapped_column(String(10))
    summary: Mapped[str] = mapped_column(Text)
    headers: Mapped[Optional[str]] = mapped_column(String(40), nullable=True)
    body: Mapped[Optional[str]] = mapped_column(String(40), nullable=True)
    created_at: Mapped[datetime] = mapped_column(TIMESTAMP(timezone=True))

    transaction_id: Mapped[str] = mapped_column(ForeignKey("transactions.id", ondelete="CASCADE"))
    transaction: Mapped["Transaction"] = relationship(back_populates="messages")

    __table_args__ = (Index("ix_transaction_id", "transaction_id"),)

    def to_model(self):
        return MessageModel(
            id=self.id,
            transaction_id=self.transaction_id,
            kind=self.kind,
            summary=self.summary,
            headers=self.headers,
            body=self.body,
            created_at=(self.created_at.replace(tzinfo=UTC) if self.created_at else self.created_at),
        )

    @classmethod
    def from_models(cls, transaction, message: BaseMessage, headers, content):
        serializer = get_serializer_for(message)

        obj = cls()
        obj.transaction_id = transaction.id
        obj.kind = message.kind
        obj.summary = serializer.summary
        obj.headers = headers.id
        obj.body = content.id
        obj.created_at = message.created_at.astimezone(UTC)
        return obj


class MessagesRepository(Repository[Message]):
    Type = Message

    @with_session
    async def create(self, values: dict | MessageModel, /, *, session):
        if isinstance(values, MessageModel):
            values = dict(
                id=values.id,
                transaction_id=values.transaction_id,
                kind=values.kind,
                summary=values.summary,
                headers=values.headers,
                body=values.body,
                created_at=values.created_at,
            )

        if not values.get("created_at"):
            values["created_at"] = datetime.now(UTC)

        return await super().create(values, session=session)



from harp.config.asdict import asdict

from harp_apps.proxy.settings import ProxySettings

short_syntax_simple_config = {
    "endpoints": [
        {
            "name": "test",
            "port": 8080,
            "url": "http://example.com",
        }
    ]
}

simple_config_with_probe = {
    "endpoints": [
        {
            "name": "test",
            "port": 8080,
            "remote": {
                "endpoints": [
                    {"url": "http://example.com"},
                    {"url": "http://fallback.example.com", "pools": ["fallback"]},
                ],
                "probe": {
                    "method": "GET",
                    "path": "/",
                    "timeout": 5,
                },
            },
        }
    ]
}


def test_short_syntax():
    settings = ProxySettings.from_dict(short_syntax_simple_config)

    assert asdict(settings) == {
        "endpoints": [
            {
                "name": "test",
                "port": 8080,
                "remote": {
                    "endpoints": [
                        {"url": "http://example.com/"},
                    ]
                },
            }
        ]
    }


def test_default_settings():
    settings = ProxySettings()
    assert asdict(settings) == {}


def test_probe_syntax():
    settings = ProxySettings.from_dict(simple_config_with_probe)
    assert asdict(settings) == {
        "endpoints": [
            {
                "name": "test",
                "port": 8080,
                "remote": {
                    "endpoints": [
                        {"url": "http://example.com/"},
                        {"pools": ["fallback"], "url": "http://fallback.example.com/"},
                    ],
                    "probe": {"timeout": 5.0},
                },
            }
        ]
    }



from typing import Annotated, Generic, Self, Type, TypeVar

from pydantic import Field, field_serializer

from .base import BaseConfigurable, Configurable

TSettings = TypeVar("TSettings", bound=Configurable)


class Stateful(BaseConfigurable, Generic[TSettings]):
    """
    A base class for stateful objects that are related to a settings class. Usually, it's used for live status of
    settings-defined objects, that may vary over time. It allows to separate the concerns of the imuatable,
    environment-provided settings, and the mutable, runtime-related state of the object.

    """

    settings: Annotated[TSettings, Field(repr=False)]

    @classmethod
    def get_settings_type(cls) -> Type[Configurable]:
        return cls.model_fields["settings"].annotation

    @classmethod
    def from_settings_dict(cls: Type[Self], data: dict) -> Self:
        settings_type = cls.get_settings_type()
        return cls(settings=settings_type(**data))

    @classmethod
    def from_settings_kwargs(cls: Type[Self], **kwargs) -> Self:
        settings_type = cls.get_settings_type()
        return cls(settings=settings_type(**kwargs))

    @field_serializer("settings", when_used="json")
    @classmethod
    def __serialize_settings(cls, settings: TSettings):
        BaseType = type(settings).__mro__[type(settings).__mro__.index(Configurable) - 1]
        return BaseType.model_construct(
            **{k: v for k, v in settings.model_dump().items() if k in BaseType.model_fields}
        ).model_dump()



from functools import cached_property

import pytest

from harp.config import Application, ApplicationsRegistry, ConfigurationBuilder
from harp.config.asdict import asdict


class BaseTestForApplications:
    name = None
    config_key = None
    expected_defaults = {}

    @cached_property
    def registry(self) -> ApplicationsRegistry:
        return ApplicationsRegistry()

    @cached_property
    def application(self) -> Application:
        return self.registry.get_application(self.name)

    @pytest.fixture
    def configuration(self):
        config = ConfigurationBuilder(use_default_applications=False)
        config.applications.add(self.name)
        return config

    def test_default_settings(self):
        """Checks that default settings are as expected by the implementation."""
        computed_defaults = asdict(self.application.settings_type(), verbose=True)
        assert computed_defaults == self.expected_defaults

    def test_default_settings_idempotence(self):
        """Checks that default settings, if used to recrete settings, will keep the same values."""
        defaults_settings = asdict(self.application.settings_type(), verbose=True)
        reparsed_settings = asdict(self.application.settings_type(**defaults_settings), verbose=True)
        assert reparsed_settings == defaults_settings



from functools import wraps
from typing import Generic, TypeVar

from sqlalchemy import delete, select, update
from sqlalchemy.exc import NoResultFound
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.sql.functions import func


class Base(AsyncAttrs, DeclarativeBase):
    pass


def with_session(f):
    """
    Decorates a method to ensure it is called with a sql alchemy session, if it is not, then create one to wrap the
    call.

    :param f:
    :return:
    """

    @wraps(f)
    async def contextualized_with_session(self, *args, session=None, **kwargs):
        if session is None:
            async with session or self.session_factory() as session:
                return await f(self, *args, session=session, **kwargs)
        return await f(self, *args, session=session, **kwargs)

    return contextualized_with_session


TRow = TypeVar("TRow")


class Repository(Generic[TRow]):
    Type: TRow = None

    def __init__(self, session_factory, /):
        self.session_factory = session_factory

    def select(self):
        return select(self.Type)

    def delete(self):
        return delete(self.Type)

    def count(self):
        return select(func.count()).select_from(self.Type)

    def update(self):
        return update(self.Type)

    @with_session
    async def find_one(self, values: dict, /, session, **select_kwargs) -> TRow:
        return (
            (
                await session.execute(
                    self.select(**select_kwargs).where(
                        *(getattr(self.Type, attr) == value for attr, value in values.items())
                    )
                )
            )
            .unique()
            .scalar_one()
        )

    @with_session
    async def find_one_by_id(self, id: str, /, session=None, **select_kwargs) -> TRow:
        return await self.find_one({"id": id}, session=session, **select_kwargs)

    @with_session
    async def find_or_create_one(self, values: dict, /, session, defaults=None, **select_kwargs) -> TRow:
        try:
            return await self.find_one(values, session=session, **select_kwargs)
        except NoResultFound:
            return await self.create((defaults or {}) | values, session=session)

    @with_session
    async def create(self, values: dict, /, *, session):
        item = self.Type(**values)
        session.add(item)
        await session.commit()
        return item



from typing import TYPE_CHECKING

from pydantic import Field

from harp.config import Service
from harp_apps.http_client.settings.cache import CacheSettings

if TYPE_CHECKING:
    pass

from harp.settings import DEFAULT_TIMEOUT


class HttpClientSettings(Service):
    type: str = Field("httpx.AsyncClient", description=Service.model_fields["type"].description)
    arguments: dict = {"timeout": DEFAULT_TIMEOUT}

    cache: CacheSettings = CacheSettings()

    #: HTTP transport to use for the client. This is usually a httpx.AsyncHTTPTransport (or subclass) instance.
    transport: Service = Service(
        type="httpx.AsyncHTTPTransport",
        arguments={
            "verify": True,
            "retries": 0,
        },
    )

    proxy_transport: Service = Service(
        type="harp_apps.http_client.transport.AsyncFilterableTransport",
    )



from datetime import UTC, datetime

from sqlalchemy import TIMESTAMP, ForeignKey, Integer, String, func
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base, Repository, with_session


class Metric(Base):
    __tablename__ = "metrics"

    id = mapped_column(Integer(), primary_key=True, unique=True, autoincrement=True)
    name = mapped_column(String(255), unique=True, nullable=False)


class MetricValue(Base):
    __tablename__ = "metric_values"

    metric_id = mapped_column(ForeignKey("metrics.id"), nullable=False, primary_key=True)
    metric: Mapped["Metric"] = relationship()
    created_at = mapped_column(TIMESTAMP(timezone=True), server_default=func.now(), primary_key=True)

    value = mapped_column(Integer())


class MetricValuesRepository(Repository[MetricValue]):
    Type = MetricValue


class MetricsRepository(Repository[Metric]):
    Type = Metric

    def __init__(self, session_factory, /):
        super().__init__(session_factory)
        self.values = MetricValuesRepository(session_factory)

    @with_session
    async def insert_values(self, values: dict, /, session):
        now = datetime.now(UTC)
        for name, value in values.items():
            metric = await self.find_or_create_one({"name": name}, session=session)
            await self.values.create(
                {"metric_id": metric.id, "value": value, "created_at": now},
                session=session,
            )



from datetime import date, datetime
from typing import Optional


def ensure_date(x) -> Optional[date]:
    if x is None:
        return None
    if isinstance(x, datetime):
        return x.date()
    if isinstance(x, date):
        return x
    return datetime.strptime(x, "%Y-%m-%d").date()


def ensure_datetime(x, tz=None) -> Optional[datetime]:
    if x is None:
        return None
    if isinstance(x, datetime):
        return x.replace(tzinfo=tz)
    if isinstance(x, date):
        return datetime.combine(x, datetime.min.time(), tzinfo=tz)
    try:
        return datetime.strptime(x, "%Y-%m-%d %H:%M:%S").replace(tzinfo=tz)
    except ValueError:
        return datetime.strptime(x, "%Y-%m-%d %H:%M:%S.%f").replace(tzinfo=tz)



"""
Errors (:mod:`harp.errors`) module contains the exception hierarchy for Harp.

Contents
--------

"""

__title__ = "Errors"


class BaseError(Exception):
    pass


class ConfigurationError(BaseError):
    pass


class ConfigurationValueError(ConfigurationError, ValueError):
    pass


class ConfigurationRuntimeError(ConfigurationError, RuntimeError):
    pass


class ConfigurationRemovedSettingError(ConfigurationError):
    pass



from typing import cast
from unittest.mock import AsyncMock

import pytest
from asgiref.typing import HTTPScope

from harp.asgi import ASGIKernel
from harp.asgi.bridge.requests import HttpRequestAsgiBridge
from harp.controllers import DefaultControllerResolver
from harp.http import HttpRequest, HttpResponse
from harp.utils.bytes import ensure_bytes
from harp.utils.testing.communicators import ASGICommunicator
from harp.views.json import register as register_json_views


async def _create_request(body=None, /, *, method="GET", headers=None, path="/"):
    receive = AsyncMock(
        return_value={"body": ensure_bytes(body) if body else b""},
    )
    request = HttpRequest(
        HttpRequestAsgiBridge(
            cast(
                HTTPScope,
                {
                    "type": "http",
                    "method": method,
                    "path": path,
                    **(
                        {"headers": [(ensure_bytes(k), ensure_bytes(v)) for k, v in headers.items()]} if headers else {}
                    ),
                },
            ),
            receive,
        ),
    )
    return request


class ControllerTestFixtureMixin:
    ControllerType = None

    def create_controller(self, *args, **kwargs):
        return self.ControllerType(*args, **kwargs)

    async def call_controller(
        self, controller=None, /, *, body=None, method="GET", headers=None
    ) -> tuple[HttpRequest, HttpResponse]:
        if controller is None:
            controller = self.create_controller()
        request = await _create_request(body or b"", method=method, headers=headers)
        return request, await controller(request)

    @pytest.fixture
    def controller(self, storage):
        raise NotImplementedError('You must implement the "controller" fixture in your test class.')


class ControllerThroughASGIFixtureMixin(ControllerTestFixtureMixin):
    @pytest.fixture
    def kernel(self, controller):
        kernel = ASGIKernel(
            resolver=DefaultControllerResolver(default_controller=controller),
            handle_errors=False,
        )
        register_json_views(kernel.dispatcher)
        return kernel

    @pytest.fixture
    async def client(self, kernel):
        client = ASGICommunicator(kernel)
        await client.asgi_lifespan_startup()
        return client



from itertools import chain
from typing import Optional


def flatten_facet_value(values: list):
    return list(
        chain(
            *map(lambda x: x.split(","), values),
        ),
    )


def str_to_float_or_none(s: str) -> Optional[float]:
    try:
        return float(s)
    except ValueError:
        return None



import re

import pytest

pytest.register_assert_rewrite("harp.utils.testing.applications")
pytest.register_assert_rewrite("harp.utils.testing.cli")
pytest.register_assert_rewrite("harp.utils.testing.config")


class RE(object):
    def __init__(self, pattern):
        self.pattern = re.compile(pattern)

    def __eq__(self, other):
        return self.pattern.match(other) is not None

    def __ne__(self, other):
        return self.pattern.match(other) is None

    def __repr__(self):
        return f"<RE {self.pattern.pattern}>"



from tempfile import NamedTemporaryFile

from harp.config.asdict import asdict
from harp_apps.dashboard.settings import BasicAuthSettings, DashboardSettings


def test_defaults():
    assert asdict(DashboardSettings()) == {}

    assert asdict(DashboardSettings(), verbose=True) == {
        "auth": None,
        "devserver": {"enabled": True, "port": None},
        "enable_ui": True,
        "port": 4080,
        "public_url": None,
    }


def test_basic_auth_defaults():
    assert asdict(BasicAuthSettings()) == {"type": "basic"}
    assert asdict(BasicAuthSettings(), verbose=True) == {
        "algorithm": "pbkdf2_sha256",
        "type": "basic",
        "users": {},
    }


def test_basic_auth_userlist():
    settings = BasicAuthSettings.from_dict({"users": {"john": "foo", "jane": "bar"}})
    assert asdict(settings) == {
        "type": "basic",
        "users": {"jane": {"password": "bar"}, "john": {"password": "foo"}},
    }

    assert asdict(settings, verbose=True) == {
        "algorithm": "pbkdf2_sha256",
        "type": "basic",
        "users": {"jane": {"password": "bar"}, "john": {"password": "foo"}},
    }


def test_basic_auth_userlist_fromfile():
    with NamedTemporaryFile("w", suffix=".yml", delete=False) as f:
        f.write("john: foo\njane: bar\n")
        f.flush()

        assert asdict(BasicAuthSettings.from_dict({"users": {"fromFile": f.name}, "algorithm": "plain"})) == {
            "algorithm": "plaintext",
            "type": "basic",
            "users": {"jane": {"password": "bar"}, "john": {"password": "foo"}},
        }


def test_no_devserver():
    settings = DashboardSettings()
    assert settings.devserver.enabled is True


def test_devserver_disable():
    settings = DashboardSettings.from_dict({"devserver": {"enabled": False}})
    assert settings.devserver.enabled is False



from .errors import HttpError
from .requests import HttpRequest
from .responses import AlreadyHandledHttpResponse, HttpResponse, JsonHttpResponse
from .serializers import HttpRequestSerializer, get_serializer_for
from .typing import BaseHttpMessage, BaseMessage, HttpRequestBridge, HttpResponseBridge

__title__ = "HTTP"

__all__ = [
    "AlreadyHandledHttpResponse",
    "BaseHttpMessage",
    "BaseMessage",
    "HttpError",
    "HttpRequest",
    "HttpRequestBridge",
    "HttpRequestSerializer",
    "HttpResponse",
    "HttpResponseBridge",
    "JsonHttpResponse",
    "get_serializer_for",
]



from decimal import Decimal

from pydantic import HttpUrl

from harp.views.json import serialize


def test_decimal_serialization():
    decimal = Decimal("3.14")
    assert serialize(decimal) == b'"3.14"'


def test_url_serialization():
    url = HttpUrl("https://example.com")
    assert serialize(url) == b'"https://example.com/"'



import orjson
from httpx import AsyncByteStream, ByteStream, codes
from multidict import CIMultiDict
from typing import Optional

from harp.utils.bytes import ensure_bytes
from .typing import BaseHttpMessage


class HttpResponse(BaseHttpMessage):
    kind = "response"

    def __init__(
        self,
        body: bytes | str,
        /,
        *,
        status: int = 200,
        headers: Optional[dict] = None,
        content_type=None,
        extensions: Optional[dict] = None,
    ):
        super().__init__(extensions=extensions)

        self._body = ensure_bytes(body)
        self._status = int(status)
        self._headers = CIMultiDict(headers or {})
        self._stream: AsyncByteStream = ByteStream(self._body)

        if content_type:
            self._headers["content-type"] = content_type

    @property
    def stream(self):
        return self._stream

    @stream.setter
    def stream(self, stream):
        self._stream = stream
        if hasattr(self, "_body"):
            delattr(self, "_body")

    @property
    def body(self) -> bytes:
        if not hasattr(self, "_body"):
            raise RuntimeError("The 'body' attribute is not available, please await `aread()` first.")
        return self._body

    @property
    def status(self) -> int:
        return self._status

    @property
    def reason_phrase(self) -> str:
        try:
            reason_phrase = self.extensions["reason_phrase"]
        except KeyError:
            return codes.get_reason_phrase(self.status)

        if isinstance(reason_phrase, bytes):
            return reason_phrase.decode("ascii", errors="ignore")
        return reason_phrase

    @property
    def headers(self) -> CIMultiDict:
        return self._headers

    @headers.setter
    def headers(self, headers: CIMultiDict):
        self._headers = CIMultiDict(headers)

    @property
    def content_type(self) -> str:
        return self._headers.get("content-type", "text/plain")

    async def aread(self):
        if not hasattr(self, "_body"):
            self._body = b"".join([part async for part in self._stream])
        if not isinstance(self._stream, ByteStream):
            self._stream = ByteStream(self._body)
        return self.body


class JsonHttpResponse(HttpResponse):
    def __init__(self, body: dict, /, *, status: int = 200, headers: dict = None):
        from harp.views.json import serialize

        super().__init__(
            serialize(body),
            status=status,
            headers=headers,
            content_type="application/json",
        )

    def asdict(self):
        return orjson.loads(self.body)


class AlreadyHandledHttpResponse(HttpResponse):
    def __init__(self):
        super().__init__(b"")



from asgiref.typing import ASGIApplication, LifespanScope, LifespanStartupEvent


class ASGICommunicator:
    default_host = "localhost"
    default_port = 80

    def __init__(self, asgi_app: ASGIApplication, *, default_host=None, default_port=None):
        self.asgi_app = asgi_app
        self.default_host = default_host or self.default_host
        self.default_port = default_port or self.default_port

    async def asgi_lifespan_startup(self):
        from asgiref.testing import ApplicationCommunicator

        com = ApplicationCommunicator(
            self.asgi_app,
            LifespanScope(
                type="lifespan",
                asgi={
                    "spec_version": "2.1",
                    "version": "3.0",
                },
            ),
        )
        await com.send_input(
            LifespanStartupEvent(type="lifespan.startup"),
        )
        await com.wait()

    def http_request(self, method, path, *, body=b"", headers=None, host=None, port=None):
        from harp.utils.testing.communicators import HTTPCommunicator

        communicator = HTTPCommunicator(
            self.asgi_app,
            method,
            path,
            body=body,
            headers=headers,
            hostname=host or self.default_host,
            port=port or self.default_port,
        )

        return communicator.get_response()

    def http_get(self, path, *, headers=None, host=None, port=None):
        return self.http_request("GET", path, headers=headers, host=host, port=port)

    def http_post(self, path, *, body=b"", headers=None, host=None, port=None):
        return self.http_request("POST", path, body=body, headers=headers, host=host, port=port)

    def http_put(self, path, *, headers=None, host=None, port=None):
        return self.http_request("PUT", path, headers=headers, host=host, port=port)

    def http_patch(self, path, *, headers=None, host=None, port=None):
        return self.http_request("PATCH", path, headers=headers, host=host, port=port)

    def http_delete(self, path, *, headers=None, host=None, port=None):
        return self.http_request("DELETE", path, headers=headers, host=host, port=port)

    def http_options(self, path, *, headers=None, host=None, port=None):
        return self.http_request("OPTIONS", path, host=host, port=port)

    def http_head(self, path, *, headers=None, host=None, port=None):
        return self.http_request("HEAD", path, headers=headers, host=host, port=port)



from harp import get_logger
from harp.config import Application, OnBindEvent, OnBoundEvent, OnReadyEvent, OnShutdownEvent
from harp_apps.acme.settings import AcmeSettings

logger = get_logger(__name__)


async def on_bind(event: OnBindEvent):
    logger.warning("🔗 Binding Acme services")


async def on_bound(event: OnBoundEvent):
    logger.warning("🔗 Bound Acme services")


async def on_ready(event: OnReadyEvent):
    logger.warning("🔗 Building Acme services")


async def on_shutdown(event: OnShutdownEvent):
    logger.warning("🔗 Disposing Acme services")


application = Application(
    on_bind=on_bind,
    on_bound=on_bound,
    on_ready=on_ready,
    on_shutdown=on_shutdown,
    settings_type=AcmeSettings,
)



from datetime import datetime

import pytest
from sqlalchemy import inspect
from sqlalchemy.exc import MissingGreenlet, StatementError
from sqlalchemy.orm.exc import DetachedInstanceError

from harp.utils.guids import generate_transaction_id_ksuid
from harp_apps.storage.services.sql import SqlStorage
from harp_apps.storage.utils.testing.mixins import StorageTestFixtureMixin


class TestModelsBase(StorageTestFixtureMixin):
    async def test_create_using_explicit_session(self, sql_storage: SqlStorage):
        """
        Check how instance creation works with an explicit session scope created outside the "create" call.

        """
        async with sql_storage.session_factory() as session:
            # we create a transaction with our own session
            db_transaction = await sql_storage.transactions.create(
                {
                    "id": generate_transaction_id_ksuid(),
                    "type": "http",
                    "endpoint": "/api/transactions",
                    "started_at": datetime.now(),
                },
                session=session,
            )

            # but lazy loading cannot work there ...
            with pytest.raises((MissingGreenlet, StatementError)):
                _ = db_transaction._tag_values

            # ... unless we use the async wrapper ...
            assert await db_transaction.awaitable_attrs._tag_values == []

            # ... and after that, the attribute is now populated
            assert db_transaction._tag_values == []

            # within the session, the object is not detached nor expired
            assert inspect(db_transaction).detached is False
            assert inspect(db_transaction).expired is False

        # although it will be dettached once the session is closed
        assert inspect(db_transaction).detached is True
        assert inspect(db_transaction).expired is False

    async def test_create_using_inplicit_session(self, sql_storage: SqlStorage):
        """
        Check how instance creation works with an implicit session.

        """
        # we create a transaction without an explicit session
        db_transaction = await sql_storage.transactions.create(
            {
                "id": generate_transaction_id_ksuid(),
                "type": "http",
                "endpoint": "/api/transactions",
                "started_at": datetime.now(),
            },
        )

        # we cannot lazy load on a detached instance ...
        with pytest.raises(DetachedInstanceError):
            _ = db_transaction._tag_values

        # ... not even with the async wrapper
        with pytest.raises(DetachedInstanceError):
            _ = await db_transaction.awaitable_attrs._tag_values

        # we can see it is (correctly) detached
        assert inspect(db_transaction).detached is True
        assert inspect(db_transaction).expired is False



"""auto-generated migration

Revision ID: 248285939c4c
Revises:
Create Date: 2024-06-09 08:22:34.388765

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "248285939c4c"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.create_table(
        "blobs",
        sa.Column("id", sa.String(length=40), nullable=False),
        sa.Column("data", sa.LargeBinary(), nullable=True),
        sa.Column("content_type", sa.String(length=64), nullable=True),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("id"),
    )
    op.create_table(
        "metrics",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_table(
        "tags",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("name", sa.String(length=64), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_table(
        "transactions",
        sa.Column("id", sa.String(length=27), nullable=False),
        sa.Column("type", sa.String(length=10), nullable=True),
        sa.Column("endpoint", sa.String(length=32), nullable=True),
        sa.Column("started_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("finished_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("elapsed", sa.Float(), nullable=True),
        sa.Column("apdex", sa.Integer(), nullable=True),
        sa.Column("x_method", sa.String(length=16), nullable=True),
        sa.Column("x_status_class", sa.String(length=3), nullable=True),
        sa.Column("x_cached", sa.String(length=32), nullable=True),
        sa.Column("x_no_cache", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("id"),
    )
    op.create_index(op.f("ix_transactions_endpoint"), "transactions", ["endpoint"], unique=False)
    op.create_index(op.f("ix_transactions_started_at"), "transactions", ["started_at"], unique=False)
    op.create_index(op.f("ix_transactions_type"), "transactions", ["type"], unique=False)
    op.create_index(op.f("ix_transactions_x_method"), "transactions", ["x_method"], unique=False)
    op.create_index(
        op.f("ix_transactions_x_status_class"),
        "transactions",
        ["x_status_class"],
        unique=False,
    )
    op.create_table(
        "users",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("username", sa.String(length=32), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("id"),
        sa.UniqueConstraint("username"),
    )
    op.create_table(
        "messages",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("kind", sa.String(length=10), nullable=True),
        sa.Column("summary", sa.String(length=255), nullable=True),
        sa.Column("headers", sa.String(length=40), nullable=True),
        sa.Column("body", sa.String(length=40), nullable=True),
        sa.Column("created_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("transaction_id", sa.String(length=27), nullable=True),
        sa.ForeignKeyConstraint(["transaction_id"], ["transactions.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("id"),
    )
    op.create_table(
        "metric_values",
        sa.Column("metric_id", sa.Integer(), nullable=False),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column("value", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["metric_id"],
            ["metrics.id"],
        ),
        sa.PrimaryKeyConstraint("metric_id", "created_at"),
    )
    op.create_table(
        "tag_values",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("tag_id", sa.Integer(), nullable=False),
        sa.Column("value", sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(["tag_id"], ["tags.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("id"),
        sa.UniqueConstraint("tag_id", "value", name="_tag_values_uc"),
    )
    op.create_table(
        "trans_user_flags",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("type", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("transaction_id", sa.String(length=27), nullable=False),
        sa.ForeignKeyConstraint(["transaction_id"], ["transactions.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["user_id"], ["users.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("id"),
        sa.UniqueConstraint("user_id", "transaction_id", "type", name="_user_transaction_uc"),
    )
    op.create_table(
        "trans_tag_values",
        sa.Column("transaction_id", sa.String(length=27), nullable=False),
        sa.Column("value_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["transaction_id"], ["transactions.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["value_id"], ["tag_values.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("transaction_id", "value_id"),
    )


def downgrade() -> None:
    op.drop_table("trans_tag_values")
    op.drop_table("trans_user_flags")
    op.drop_table("tag_values")
    op.drop_table("metric_values")
    op.drop_table("messages")
    op.drop_table("users")
    op.drop_index(op.f("ix_transactions_x_status_class"), table_name="transactions")
    op.drop_index(op.f("ix_transactions_x_method"), table_name="transactions")
    op.drop_index(op.f("ix_transactions_type"), table_name="transactions")
    op.drop_index(op.f("ix_transactions_started_at"), table_name="transactions")
    op.drop_index(op.f("ix_transactions_endpoint"), table_name="transactions")
    op.drop_table("transactions")
    op.drop_table("tags")
    op.drop_table("metrics")
    op.drop_table("blobs")



from typing import Optional

from httpx import AsyncClient


class GoogleChatNotificationSender:
    def __init__(self, webhook_url: str, public_url: Optional[str] = None) -> None:
        self.webhook_url = webhook_url
        self.public_url = public_url

    async def send_notification(
        self,
        method: Optional[str],
        url: Optional[str],
        status_code: int,
        message: str,
        transaction_id: Optional[str],
    ) -> None:
        error_message = self._format_error_message(method, url, status_code, message, transaction_id, self.public_url)
        async with AsyncClient() as client:
            response = await client.post(self.webhook_url, json=error_message)
            response.raise_for_status()
            assert response.status_code == 200

    @staticmethod
    def _format_error_message(
        method: Optional[str],
        url: Optional[str],
        status_code: int,
        message: str,
        transaction_id: Optional[str],
        public_url: Optional[str] = None,
    ) -> dict:
        sections = [
            {"widgets": [{"textParagraph": {"text": f"<b>Error in {method} request to {url}</b>"}}]},
            {
                "widgets": [
                    {
                        "keyValue": {
                            "topLabel": "Error Code",
                            "content": f"<b>{status_code}</b>",
                        }
                    },
                    {"keyValue": {"topLabel": "Status", "content": message}},
                ]
            },
            {"widgets": [{"textParagraph": {"text": f"<i>Transaction ID:</i> <b>{transaction_id}</b>"}}]},
        ]

        if public_url:
            sections.append(
                {
                    "widgets": [
                        {
                            "buttons": [
                                {
                                    "textButton": {
                                        "text": "VIEW LOGS",
                                        "onClick": {"openLink": {"url": f"{public_url}/transactions/{transaction_id}"}},
                                    }
                                }
                            ]  # type: ignore
                        }
                    ]
                }
            )
        return {
            "cards": [
                {
                    "header": {
                        "title": "🔴 Error Notification",
                    },
                    "sections": sections,
                }
            ]
        }



import sys

from harp.utils.processes import check_output


def parse_dependencies(dependencies: list[str]) -> dict[str, str]:
    """Transform a list of dependencies into a dictionary of dependencies, with package name as key."""
    parsed = {}

    for dependency in dependencies:
        # Skip empty lines or comments
        if not dependency or dependency.startswith("#"):
            continue

        # Handle editable packages
        if dependency.startswith("-e"):
            pkg_repo, pkg_name = dependency.rsplit("/", 1)
            pkg_name = pkg_name.split("#egg=")[-1]  # Extract package name
            pkg_version = "editable"
        else:
            parts = dependency.split("==")
            if len(parts) == 2:
                pkg_name, pkg_version = parts
            else:
                # Handle packages without a version or with unusual formatting
                pkg_name, pkg_version = parts[0], "unknown"

        # Check for duplicates
        if pkg_name in parsed:
            raise ValueError(f"Duplicate package name {pkg_name} found.")
        parsed[pkg_name] = pkg_version
    return parsed


async def get_python_dependencies():
    return list(
        filter(
            lambda x: x and not x.startswith("#"),
            (await check_output(sys.executable, "-m", "pip", "freeze")).decode("utf-8").split("\n"),
        ),
    )



from datetime import UTC
from functools import partial
from math import log10

from sqlalchemy import insert, update
from sqlalchemy.ext.asyncio import AsyncEngine
from whistle import IAsyncEventDispatcher

from harp import get_logger
from harp.http import get_serializer_for
from harp.models import Blob
from harp.utils.background import AsyncWorkerQueue
from harp_apps.proxy.events import (
    EVENT_TRANSACTION_ENDED,
    EVENT_TRANSACTION_MESSAGE,
    EVENT_TRANSACTION_STARTED,
    HttpMessageEvent,
    TransactionEvent,
)
from harp_apps.storage.constants import (
    SKIP_REQUEST_BODY_STORAGE,
    SKIP_REQUEST_HEADERS_STORAGE,
    SKIP_REQUEST_STORAGE,
    SKIP_RESPONSE_BODY_STORAGE,
    SKIP_RESPONSE_HEADERS_STORAGE,
    SKIP_RESPONSE_STORAGE,
    SKIP_STORAGE,
)
from harp_apps.storage.models import Message as SqlMessage
from harp_apps.storage.models import Transaction as SqlTransaction
from harp_apps.storage.types import IBlobStorage, IStorage

logger = get_logger(__name__)


class StorageAsyncWorkerQueue(AsyncWorkerQueue):
    def __init__(self, engine: AsyncEngine, storage: IStorage, blob_storage: IBlobStorage):
        self.engine = engine
        self.storage = storage
        self.blob_storage = blob_storage
        super().__init__()
        self.seen = set()

    def register_events(self, dispatcher: IAsyncEventDispatcher):
        dispatcher.add_listener(EVENT_TRANSACTION_STARTED, self.on_transaction_started)
        dispatcher.add_listener(EVENT_TRANSACTION_MESSAGE, self.on_transaction_message)
        dispatcher.add_listener(EVENT_TRANSACTION_ENDED, self.on_transaction_ended)

    def cleanup(self):
        super().cleanup()

    @property
    def pressure(self):
        if self._pressure <= 1:
            return 0
        return int(log10(self._pressure))

    async def on_transaction_started(self, event: TransactionEvent):
        """Event handler to store the transaction in the database."""

        if self.pressure >= 4:
            event.transaction.markers.add(SKIP_STORAGE)
            return

        # Copy fields into a dict that won't change before the task is handled
        transaction_data = event.transaction.as_storable_dict(with_tags=True)

        async def create_transaction():
            return await self.storage.transactions.create(transaction_data)

        # Schedule
        await self.push(create_transaction)

    async def on_transaction_message(self, event: HttpMessageEvent):
        if SKIP_STORAGE in event.transaction.markers or self.pressure >= 3:
            return

        # Check if we should skip this entire message
        skip_message = (
            event.message.kind == "request"
            and SKIP_REQUEST_STORAGE in event.transaction.markers
            or event.message.kind == "response"
            and SKIP_RESPONSE_STORAGE in event.transaction.markers
        )
        if skip_message:
            logger.debug(
                f"Skipping storing {event.message.kind.capitalize()} message",
                transaction_id=event.transaction.id,
            )
            return

        await event.message.aread()
        serializer = get_serializer_for(event.message)

        message_data = {
            "transaction_id": event.transaction.id,
            "kind": event.message.kind,
            "summary": serializer.summary,
            "created_at": event.message.created_at,
        }

        # Check if we should skip headers
        skip_headers = (
            event.message.kind == "request"
            and SKIP_REQUEST_HEADERS_STORAGE in event.transaction.markers
            or event.message.kind == "response"
            and SKIP_RESPONSE_HEADERS_STORAGE in event.transaction.markers
        )

        # Eventually store the headers blob (later)
        if not skip_headers and self.pressure <= 2:
            headers_blob = Blob.from_data(serializer.headers, content_type="http/headers")
            await self.push(partial(self.blob_storage.put, headers_blob), ignore_errors=True)
            message_data["headers"] = headers_blob.id
        elif skip_headers:
            logger.debug(
                f"Skipping storing {event.message.kind.capitalize()} headers",
                transaction_id=event.transaction.id,
                summary=serializer.summary,
            )

        # Check if we should skip body
        skip_body = (
            event.message.kind == "request"
            and SKIP_REQUEST_BODY_STORAGE in event.transaction.markers
            or event.message.kind == "response"
            and SKIP_RESPONSE_BODY_STORAGE in event.transaction.markers
        )

        # Eventually store the content blob (later)
        if not skip_body and self.pressure <= 1:
            content_blob = Blob.from_data(serializer.body, content_type=event.message.headers.get("content-type"))
            await self.push(partial(self.blob_storage.put, content_blob), ignore_errors=True)
            message_data["body"] = content_blob.id
        elif skip_body:
            logger.debug(
                f"Skipping storing {event.message.kind.capitalize()} body",
                transaction_id=event.transaction.id,
                summary=serializer.summary,
            )

        async def create_message():
            async with self.engine.connect() as conn:
                await conn.execute(insert(SqlMessage).values(**message_data))
                await conn.commit()

        await self.push(create_message)

    async def on_transaction_ended(self, event: TransactionEvent):
        if SKIP_STORAGE in event.transaction.markers:
            return

        transaction_id = event.transaction.id
        transaction_data = {
            "finished_at": event.transaction.finished_at.astimezone(UTC),
            "elapsed": event.transaction.elapsed,
            "tpdex": event.transaction.tpdex,
            "x_status_class": event.transaction.extras.get("status_class"),
            "x_cached": event.transaction.extras.get("cached"),
        }

        async def update_transaction():
            async with self.engine.connect() as conn:
                await conn.execute(
                    update(SqlTransaction)
                    .where(
                        SqlTransaction.id == transaction_id,
                    )
                    .values(**transaction_data)
                )
                await conn.commit()

        await self.push(update_transaction)



from multidict import MultiDict

from harp_apps.dashboard.filters.base import AbstractMinMaxFacet, NonExhaustiveFacet


def test_non_exhaustive_facet():
    facet = NonExhaustiveFacet()
    assert "NULL" in facet.choices


class TestAbstractMinMaxFacet:
    facet = AbstractMinMaxFacet()

    def test_values(self):
        self.facet.meta = {"min": 10, "max": 90}
        assert self.facet.values == {"min": 10, "max": 90}

    def test_filter(self):
        result = self.facet.filter(10, 90)
        assert result == {"values": ["min", "max"], "current": {"min": 10, "max": 90}}

    def test_filter_from_query(self):
        self.facet.name = "test"
        query = MultiDict({"testmin": "10", "testmax": "90"})
        result = self.facet.filter_from_query(query)
        assert result == {
            "values": ["min", "max"],
            "current": {"min": 10.0, "max": 90.0},
        }

    def test_get_filter(self):
        result = self.facet.get_filter(10, 90)
        assert result == {"min": 10, "max": 90}

    def test_get_filter_from_query(self):
        self.facet.name = "test"
        query = MultiDict({"testmin": "10", "testmax": "90"})
        result = self.facet.get_filter_from_query(query)
        assert result == {"min": 10.0, "max": 90.0}

    def test_min_max_from_query(self):
        self.facet.name = "test"
        query = MultiDict({"testmin": "10", "testmax": "90"})
        result = self.facet._min_max_from_query(query)
        assert result == (10.0, 90.0)



from urllib.parse import unquote, urlparse

from asgiref.testing import ApplicationCommunicator


class HTTPCommunicator(ApplicationCommunicator):
    """
    ApplicationCommunicator subclass that has HTTP shortcut methods.

    It will construct the scope for you, so you need to pass the application
    (uninstantiated) along with HTTP parameters.

    This does not support full chunking - for that, just use ApplicationCommunicator
    directly.
    """

    def __init__(self, application, method, path, body=b"", headers=None, *, hostname="0.0.0.0", port=80):
        parsed = urlparse(path)
        self.scope = {
            "type": "http",
            "http_version": "1.1",
            "method": method.upper(),
            "path": unquote(parsed.path),
            "query_string": parsed.query.encode("utf-8"),
            "headers": headers or [],
            **({"server": (hostname, port)} if port and hostname else {}),
        }
        assert isinstance(body, bytes)
        self.body = body
        self.sent_request = False
        super().__init__(application, self.scope)

    async def get_response(self, timeout=5):
        """
        Get the application's response. Returns a dict with keys of
        "body", "headers" and "status".
        """
        # If we've not sent the request yet, do so
        if not self.sent_request:
            self.sent_request = True
            await self.send_input({"type": "http.request", "body": self.body})
        # Get the response start
        response_start = await self.receive_output(timeout)
        assert response_start["type"] == "http.response.start"
        # Get all body parts
        response_start["body"] = b""
        while True:
            chunk = await self.receive_output(timeout)
            assert chunk["type"] == "http.response.body"
            assert isinstance(chunk["body"], bytes)
            response_start["body"] += chunk["body"]
            if not chunk.get("more_body", False):
                break
        # Return structured info
        del response_start["type"]
        response_start.setdefault("headers", [])
        return response_start



from typing import override

from harp.models import Blob
from harp_apps.storage.types import IBlobStorage


class NullBlobStorage(IBlobStorage):
    type = "null"

    async def get(self, blob_id: str):
        return None

    @override
    async def put(self, blob: Blob) -> Blob:
        return blob

    force_put = put

    async def delete(self, blob_id: str):
        pass

    async def exists(self, blob_id: str) -> bool:
        return False



from typing import Annotated

from pydantic import AfterValidator, BeforeValidator, UrlConstraints
from pydantic_core import MultiHostUrl
from sqlalchemy import URL

from harp.config import Configurable

ALLOWED_SCHEMES = [
    "mysql",
    "mysql+aiomysql",
    "mysql+asyncmy",
    "postgresql",
    "postgresql+asyncpg",
    "sqlite",
    "sqlite+aiosqlite",
]


def _unwrap_sqlalchemy_urls(url):
    if isinstance(url, URL):
        return url.render_as_string(hide_password=False)
    return url


def _fix_unasync_dsns(url: str | MultiHostUrl) -> MultiHostUrl:
    if isinstance(url, MultiHostUrl):
        url = str(url)
    if url.startswith("mysql://"):
        url = url.replace("mysql://", "mysql+aiomysql://")
    elif url.startswith("postgresql://"):
        url = url.replace("postgresql://", "postgresql+asyncpg://")
    elif url.startswith("sqlite://"):
        url = url.replace("sqlite://", "sqlite+aiosqlite://")
    return MultiHostUrl(url)


DatabaseUrl = Annotated[
    MultiHostUrl,
    UrlConstraints(allowed_schemes=ALLOWED_SCHEMES),
    BeforeValidator(_unwrap_sqlalchemy_urls),
    AfterValidator(_fix_unasync_dsns),
]


class DatabaseSettings(Configurable):
    url: DatabaseUrl = DatabaseUrl("sqlite+aiosqlite:///:memory:?cache=shared")



import asyncio
from concurrent.futures import ThreadPoolExecutor
from contextlib import asynccontextmanager
from typing import Union

from pydantic_core import MultiHostUrl, Url
from sqlalchemy import URL, make_url, text
from sqlalchemy.ext.asyncio import create_async_engine

from harp.commandline import migrations
from harp.utils.testing.cli import CliRunner


async def run_cli_migrate_command(url: Union[str | URL | Url | MultiHostUrl], /, *, operation="up", revision="head"):
    """
    Run the `migrate` command from the CLI, synchronously, in a thread (via ThreadPoolExecutor). This uses click's
    CliRunner which is a testing helper. You should not use this outside tests, as it will setup a fully separate harp
    instance with its own configuration, etc.

    :param url:
    :param operation: up or down
    :param revision: target revision ("head" for latest, "base" for initial or specific revision number)
    :return:
    """
    if isinstance(url, (Url, MultiHostUrl)):
        url = str(url)

    url = make_url(url)

    def _migrate():
        cli = CliRunner()
        return cli.invoke(
            migrations.migrate,
            [
                operation,
                revision,
                "--set",
                "storage.url=" + url.render_as_string(hide_password=False),
            ],
        )

    with ThreadPoolExecutor() as executor:
        result = await asyncio.get_event_loop().run_in_executor(executor, _migrate)

    return result


@asynccontextmanager
async def get_scoped_database_url(database_url, test_id):
    """
    Returns a "scoped" database URL that is unique to the test_id, with a database available to run some tests. Logic
    differs a bit depending on the target dialect, but the idea is to create a new database for each test, and drop it
    afterward. For sqlite, we bypass this as the "migration" process is a simple reset+recreate all.

    :param database_url:
    :param test_id:
    :return:
    """
    admin_url = make_url(database_url)
    dialect = admin_url.get_dialect().name

    if dialect == "mysql":
        admin_url = admin_url.set(username="root")

    admin_engine = create_async_engine(admin_url.render_as_string(hide_password=False))

    if dialect == "sqlite":
        try:
            yield database_url
        finally:
            await admin_engine.dispose()

    elif dialect in ("mysql", "postgresql"):
        db_name = f"test_{test_id}"
        url = make_url(database_url).set(database=db_name)

        async with admin_engine.connect() as conn:
            await conn.execution_options(isolation_level="AUTOCOMMIT")

            await conn.execute(text(f"CREATE DATABASE {url.database}"))
            if dialect == "mysql":
                await conn.execute(text(f"GRANT ALL PRIVILEGES ON {url.database}.* TO '{url.username}'@'%'"))

        try:
            yield url.render_as_string(hide_password=False)
        finally:
            async with admin_engine.connect() as conn:
                await conn.execution_options(isolation_level="AUTOCOMMIT")
                await conn.execute(text(f"DROP DATABASE {url.database}"))
            await admin_engine.dispose()

    else:
        raise RuntimeError(f"Unsupported dialect «{admin_engine.dialect.name}»")


def create_database_container_for(dialect, image, driver):
    """
    Create a database container for the given dialect, image and driver.
    This is a generator that yields the connection url (DSN) for the database container, once it is ready to accept
    connections.

    It's primarily used to write pytest fixtures that connects to this containairized database.

    :param dialect:
    :param image:
    :param driver:

    :return: str
    """
    if dialect == "sqlite":
        yield f"{dialect}+{driver}:///:memory:"
    elif dialect == "postgresql":
        from testcontainers.postgres import PostgresContainer

        with PostgresContainer(image) as container:
            yield container.get_connection_url().replace("postgresql+psycopg2://", f"{dialect}+{driver}://")
    elif dialect == "mysql":
        from testcontainers.mysql import MySqlContainer

        with MySqlContainer(image) as container:
            yield container.get_connection_url().replace("mysql+pymysql://", f"{dialect}+{driver}://")
    elif dialect == "mssql":
        from testcontainers.mssql import SqlServerContainer

        with SqlServerContainer() as container:
            yield container.get_connection_url().replace("mssql+pymssql://", f"{dialect}+{driver}://")
    else:
        raise ValueError(f"Unsupported or invalid dialect: {dialect!r}")



from typing import Literal, Optional, Protocol, TypeVar

from pydantic import model_validator

from harp.config import Configurable, Stateful
from harp_apps.proxy.constants import DOWN, UP


class BaseLivenessSettings(Configurable):
    type: Literal["ignore", "inherit", "naive", "leaky"] = "inherit"

    @model_validator(mode="before")
    @classmethod
    def __initialize_type(cls, value):
        _args = cls.model_fields["type"].annotation.__args__
        if len(_args) == 1:
            value.setdefault("type", _args[0])
        return value


class LivenessSubject(Protocol):
    status: int
    failure_reasons: Optional[set]


TSettings = TypeVar("TSettings", bound=Configurable)


class BaseLiveness(Stateful[TSettings]):
    def success(self, subject: LivenessSubject) -> bool:
        raise NotImplementedError()

    def failure(self, subject: LivenessSubject, reason: Optional[str] = None) -> bool:
        raise NotImplementedError()

    def reset_failure_reasons(self, subject: LivenessSubject):
        subject.failure_reasons = None

    def add_failure_reason(self, subject: LivenessSubject, reason: Optional[str] = None):
        if subject.failure_reasons is None:
            subject.failure_reasons = set()
        if reason:
            subject.failure_reasons.add(reason)

    def set_status_as_up_if_necessary(self, subject: LivenessSubject) -> bool:
        """Change the status to UP if it is not already UP, returns whether the state changed."""
        if subject.status != UP:
            self.reset_failure_reasons(subject)
            subject.status = UP
            return True
        return False

    def set_status_as_down_if_necessary(self, subject: LivenessSubject) -> bool:
        """Change the status to DOWN if it is not already DOWN, returns whether the state changed."""
        if subject.status != DOWN:
            subject.status = DOWN
            return True
        return False



import json
from typing import TYPE_CHECKING

from harp.asgi.events import RequestEvent
from harp.http import HttpResponse
from harp.utils.json import BytesEncoder

if TYPE_CHECKING:
    from harp.http import HttpRequest


async def dump_request_controller(request: "HttpRequest"):
    """
    TODO: use orjson ?
    """
    try:
        scope = request._impl.asgi_scope
    except AttributeError:
        scope = {}

    return HttpResponse(json.dumps(scope, cls=BytesEncoder), content_type="application/json")


async def not_found_controller():
    return HttpResponse("Not found.", status=404, content_type="text/plain")


async def ok_controller():
    return HttpResponse("Ok.", status=200)


async def on_health_request(event: RequestEvent):
    if event.request.path == "/healthz":
        event.set_controller(ok_controller)
        event.stop_propagation()



import click

from harp.config.examples import get_example_filename


def load_ruleset_from_files(files, examples, options):
    from harp_apps.rules.settings import RulesSettings

    # todo not easy to implement for now, and not really important for this command set
    if len(options):
        raise click.UsageError("Unsupported options: " + ", ".join(options))

    settings = RulesSettings()
    for file in files:
        settings.load(file)
    for example in examples:
        settings.load(get_example_filename(example))

    return settings.ruleset



from typing import Callable, Type, TypeVar, cast

T = TypeVar("T")


def factory(t: Type[T]) -> Callable[..., Type[T]]:
    def decorator(f) -> Type[T]:
        return cast(Type[T], type(f.__name__, (t,), {"__new__": f, "__init__": f}))

    return decorator



import asyncio
import dataclasses
import socket
import threading
import time
from contextlib import contextmanager
from typing import Awaitable, Callable, Optional

import httpx


def get_available_network_port():
    import socket

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def wait_for_port(port: int, host: str = "localhost", timeout: float = 10.0):
    """Wait until a port starts accepting TCP connections.
    Args:
        port: Port number.
        host: Host address on which the port should exist.
        timeout: In seconds. How long to wait before raising errors.
    Raises:
        TimeoutError: The port isn't accepting connection after time specified in `timeout`.
    """
    start_time = time.perf_counter()
    while True:
        try:
            with socket.create_connection((host, port), timeout=timeout):
                break
        except OSError as ex:
            time.sleep(0.01)
            if time.perf_counter() - start_time >= timeout:
                raise TimeoutError(
                    "Waited too long for the port {} on host {} to start accepting connections.".format(port, host)
                ) from ex


class PortReservationManager:
    """Manages port reservations to prevent race conditions."""

    def __init__(self):
        self._reserved_ports = {}
        self._lock = threading.Lock()

    def reserve_port(self) -> int:
        """Reserve a port and keep the socket open until released."""
        with self._lock:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.bind(("127.0.0.1", 0))
            port = sock.getsockname()[1]
            self._reserved_ports[port] = sock
            return port

    def release_port(self, port: int):
        """Release a reserved port."""
        with self._lock:
            if port in self._reserved_ports:
                self._reserved_ports[port].close()
                del self._reserved_ports[port]

    @contextmanager
    def reserve_port_context(self):
        """Context manager for port reservation."""
        port = self.reserve_port()
        try:
            yield port
        finally:
            self.release_port(port)


# Global port manager instance
_port_manager = PortReservationManager()


def get_reserved_port() -> int:
    """Get an available port with reservation to prevent race conditions."""
    return _port_manager.reserve_port()


async def wait_for_service_ready(
    port: int,
    host: str = "localhost",
    health_path: Optional[str] = None,
    custom_check: Optional[Callable[[int], Awaitable[bool]]] = None,
    timeout: float = 10.0,
    retry_interval: float = 0.1,
):
    """Wait for a service to be ready by checking health endpoint or custom check.

    Args:
        port: Port number.
        host: Host address.
        health_path: Optional health check endpoint path.
        custom_check: Optional custom readiness check function.
        timeout: Total timeout in seconds.
        retry_interval: Time between retries in seconds.

    Raises:
        TimeoutError: If service is not ready within timeout.
    """
    start_time = time.perf_counter()
    base_url = f"http://{host}:{port}"

    while time.perf_counter() - start_time < timeout:
        try:
            if custom_check:
                if await custom_check(port):
                    return
            elif health_path:
                async with httpx.AsyncClient() as client:
                    response = await client.get(f"{base_url}{health_path}")
                    if response.status_code == 200:
                        return
            else:
                # Fall back to simple port check
                with socket.create_connection((host, port), timeout=1):
                    return
        except Exception:
            await asyncio.sleep(retry_interval)

    raise TimeoutError(f"Service not ready on {host}:{port} after {timeout} seconds")


@dataclasses.dataclass(frozen=True)
class Bind:
    host: str
    port: int

    def __str__(self):
        return f"{self.host}:{self.port}"



"""
The Command Line (:mod:`harp.commandline`) package bundles everything related to the `harp` command.

It contains an :func:`entrypoint` callable, built using `click`, which is called when you type `harp` in your terminal.

It is a high-level development helper: for example, `harp start` wraps `honcho`, a python process manager (foreman
clone) that can spawn multiple processes to help you with development.

Production environment will favour the lower-level `bin/entrypoint` script, that skips the process manager entirely
and runs the application server directly.

This may be subject to changes in the future (especially to align/refactor arg parsers), but this approach works
well for now.

.. todo::

    @click.option('--prod/--dev', default=False, help="Set defaults for production or development environment (default
    to dev).")

    @click.option('--verbose', '-v', count=True)
    --verbose : change default log verbosity

    @click.option('--debug', is_flag=True)
    --debug : ?

    Manage process list / options using a class (with hierarchy for prod/dev, for example)
    If one process only, no honcho ? Or mayne no honcho manager ?

.. todo::

    Possible future things (or not)

    - build (compile stuff for production)
    - serve (run a production-like server)

    Document how harp start work (honcho, devserver port, ...)

"""

from typing import cast

from click import Command

from harp.commandline.config import config
from harp.commandline.create import create
from harp.commandline.examples import entrypoint as examples
from harp.commandline.server import server
from harp.commandline.version import version
from harp.settings import HARP_ENV
from harp.utils.commandline import check_packages, click, code

__title__ = "Command Line"

IS_DEVELOPMENT_ENVIRONMENT = False

if HARP_ENV == "dev" or check_packages("honcho", "watchfiles"):
    IS_DEVELOPMENT_ENVIRONMENT = True

if HARP_ENV == "prod":
    IS_DEVELOPMENT_ENVIRONMENT = False


@click.group()
def entrypoint():
    """HTTP Application Runtime Proxy (HARP)

    The following commands are available to help you setup and run your HTTP proxy application.

    """
    pass


if IS_DEVELOPMENT_ENVIRONMENT:
    from harp.commandline.start import start

    entrypoint.add_command(start)

    from harp.commandline.install import install_dev

    entrypoint.add_command(install_dev)
else:

    @click.command(
        short_help="Starts the local development environment.",
        help=f"""Starts the local development environment, using honcho to spawn a configurable set of processes that you
        can adapt to your needs. By default, it will starts the `dashboard` (frontend dev server) and `server` (python
        server) processes. For live instances, you'll prefer {code("harp server")}.""",
    )
    def start(*args, **kwargs):
        raise NotImplementedError(
            "This command is not available in production environment, please install the dev extra if you need it."
        )

    entrypoint.add_command(start)


if check_packages("alembic"):
    from harp.commandline.migrations import create_migration, feature, history, migrate, reset, run_db_merge_command

    entrypoint.add_command(migrate)
    entrypoint.add_command(feature)
    entrypoint.add_command(history)
    entrypoint.add_command(reset)
    entrypoint.add_command(run_db_merge_command)

    if IS_DEVELOPMENT_ENVIRONMENT:
        entrypoint.add_command(create_migration)

if check_packages("harp_apps.rules.commandline"):
    from harp_apps.rules.commandline import entrypoint as rules_entrypoint

    entrypoint.add_command(cast(Command, rules_entrypoint))

entrypoint.add_command(server)
entrypoint.add_command(config)
entrypoint.add_command(examples)
entrypoint.add_command(create)
entrypoint.add_command(version)

__all__ = [
    "entrypoint",
]



import asyncio
import hashlib
import platform
import sys
from functools import cached_property

from httpx import AsyncClient, TransportError
from whistle import IAsyncEventDispatcher

from harp import __version__, get_logger
from harp.asgi.events import EVENT_CORE_STARTED
from harp.typing import GlobalSettings
from harp_apps.storage.types import IStorage

logger = get_logger(__name__)


class TelemetryManager:
    default_period = 86400  # 24 hours
    default_endpoint = "https://connect.makersquad.fr/t/a"

    def __init__(
        self,
        global_settings: GlobalSettings,
        client: AsyncClient,
        storage: IStorage = None,
        dispatcher: IAsyncEventDispatcher = None,
        **kwargs,
    ):
        self.global_settings = global_settings
        self.client = client
        self.storage = storage
        self.endpoint = kwargs.pop("endpoint", None) or self.default_endpoint

        self.worker = None
        self.count = 0

        self._platform = " ".join((platform.python_implementation(), sys.version, "on", platform.platform()))
        self._host = "; ".join(platform.uname())
        self._hashed = hashlib.sha1("\n".join([self._platform, self._host]).encode("utf-8")).hexdigest()

        if dispatcher:
            dispatcher.add_listener(EVENT_CORE_STARTED, self.start, priority=-10)

    @cached_property
    def period(self):
        return self.default_period

    @cached_property
    def applications(self):
        return ",".join(map(lambda x: x.split(".")[-1], self.global_settings.get("applications", [])))

    async def start(self, event):
        self.worker = asyncio.create_task(self.loop_forever())

    async def ping(self):
        try:
            return await self.client.post(
                self.endpoint,
                json={
                    # anonymous fingerprint of instance
                    "f": self._hashed,
                    # configured applications
                    "a": self.applications,
                    # application version
                    "v": __version__,
                    # transaction count in last period (not implemented yet)
                    "c": (await self.storage.get_usage()) if self.storage else 0,
                    # activity type
                    "t": "ping" if self.count else "start",
                    # incrementing counter
                    "i": self.count,
                },
                timeout=5.0,
            )
        finally:
            self.count += 1

    async def loop_forever(self):
        while True:
            try:
                await self.ping()
            except TransportError as exc:
                logger.warning("Failed to send activity ping: %s", exc)

            await asyncio.sleep(self.period)



from functools import cached_property
from typing import TYPE_CHECKING, Any, Callable, Optional

from whistle import Event

from harp.http import HttpRequest

if TYPE_CHECKING:
    from harp.http import HttpResponse

#: Event fired when the ASGI Kernel receives a "lifespan" ASGI message.
EVENT_CORE_STARTED = "core.started"


class RequestEvent(Event):
    """
    Request-aware event, that can resolve a controller from a request.
    """

    name = "core.request"

    def __init__(self, request: HttpRequest):
        self._request = request
        self._controller = None

    @cached_property
    def request(self) -> HttpRequest:
        return self._request

    @property
    def controller(self) -> Optional[Callable]:
        return self._controller

    def set_controller(self, controller: Optional[Callable]):
        self._controller = controller


#: Event fired when the ASGI Kernel receives an HTTP request, before it's processed.
EVENT_CORE_REQUEST = RequestEvent.name


class ControllerEvent(RequestEvent):
    """
    Controller and request aware event, for controller filters (e.g. decorators...).
    """

    name = "core.controller"

    def __init__(self, request: HttpRequest, controller):
        super().__init__(request)
        self._controller = controller


#: Event fired when the ASGI Kernel has resolved a controller for a request, to allow filtering
EVENT_CORE_CONTROLLER = ControllerEvent.name


class ResponseEvent(RequestEvent):
    name = "core.response"

    def __init__(self, request: HttpRequest, response: "HttpResponse"):
        super().__init__(request)
        self.response = response


#: Event fired when the ASGI Kernel has generated a response for a request, before it's sent, allowing to filter it.
EVENT_CORE_RESPONSE = ResponseEvent.name


class ViewEvent(RequestEvent):
    """
    The view event allows to transform controller return values into response objects.
    """

    name = "core.view"

    value: Any
    response: Optional["HttpResponse"]

    def __init__(self, request: HttpRequest, value: Any):
        super().__init__(request)

        self.value = value
        self.response = None

    def set_response(self, response: "HttpResponse"):
        self.response = response
        self.stop_propagation()


#: Event fired when the ASGI Kernel has called the resolved controller but got a non-HttpResponse return value, to allow
#: transforming it into a response.
EVENT_CORE_VIEW = ViewEvent.name



from typing import Optional, cast
from unittest.mock import ANY, AsyncMock, Mock, patch

import pytest
import respx
from httpx import AsyncClient, Response
from sqlalchemy.ext.asyncio import AsyncEngine
from whistle import AsyncEventDispatcher, IAsyncEventDispatcher

from harp.config.asdict import asdict
from harp.http import HttpRequest, HttpResponse
from harp.utils.bytes import ensure_bytes
from harp.utils.testing.mixins import ControllerTestFixtureMixin
from harp.utils.testing.mixins.controllers import _create_request
from harp_apps.proxy.controllers import HttpProxyController
from harp_apps.proxy.events import EVENT_TRANSACTION_STARTED
from harp_apps.proxy.settings.remote import Remote
from harp_apps.storage.services.sql import SqlStorage
from harp_apps.storage.types import IBlobStorage, IStorage
from harp_apps.storage.utils.testing.mixins import StorageTestFixtureMixin
from harp_apps.storage.worker import StorageAsyncWorkerQueue


class DispatcherTestFixtureMixin:
    @pytest.fixture(scope="function")
    def dispatcher(self):
        return AsyncEventDispatcher()


class HttpProxyControllerTestFixtureMixin(ControllerTestFixtureMixin):
    ControllerType = HttpProxyController

    @pytest.fixture(autouse=True)
    def setup(self):
        # forces the user agent to be a known value, even if versions are incremented
        with patch("harp_apps.proxy.adapters.HttpClientProxyAdapter.user_agent", "test/1.0"):
            yield

    def mock_http_endpoint(self, url, /, *, status=200, content=""):
        """Make sure you decorate your tests function using this with respx.mock decorator, otherwise the real network
        will be called and you may have some headaches..."""
        return respx.get(url).mock(return_value=Response(status, content=ensure_bytes(content)))

    def create_controller(
        self,
        url=None,
        *args,
        dispatcher: Optional[IAsyncEventDispatcher] = None,
        http_client=None,
        **kwargs,
    ):
        return super().create_controller(
            Remote.from_settings_dict({"endpoints": [{"url": url or "http://example.com/"}]}),
            *args,
            dispatcher=dispatcher,
            http_client=http_client or AsyncClient(),
            **kwargs,
        )

    def create_worker(
        self,
        dispatcher: IAsyncEventDispatcher,
        engine: AsyncEngine,
        sql_storage: IStorage,
        blob_storage: IBlobStorage,
    ) -> StorageAsyncWorkerQueue:
        worker = StorageAsyncWorkerQueue(engine, sql_storage, blob_storage)
        worker.register_events(dispatcher)
        return worker


class TestHttpProxyController(HttpProxyControllerTestFixtureMixin, DispatcherTestFixtureMixin):
    @respx.mock
    async def test_basic_get(self, dispatcher: IAsyncEventDispatcher):
        endpoint = self.mock_http_endpoint("http://example.com/", content="Hello.")
        request, response = await self.call_controller(
            self.create_controller("http://example.com/", dispatcher=dispatcher)
        )

        # check output and side effects
        assert endpoint.called and endpoint.call_count == 1
        assert response.status == 200
        assert response.headers == {}
        assert response.body == b"Hello."

    @respx.mock
    async def test_get_with_tags(self, dispatcher: IAsyncEventDispatcher):
        endpoint = self.mock_http_endpoint("http://example.com/", content="Hello.")

        # register a mock handler to inspect the actually created transaction
        transaction_started_handler = AsyncMock()
        dispatcher.add_listener(EVENT_TRANSACTION_STARTED, transaction_started_handler)

        # call our controller
        request, response = await self.call_controller(
            self.create_controller("http://example.com/", dispatcher=dispatcher),
            headers={
                "x-harp-foo": "bar",
                "accept": "application/json",
                "vary": "custom",
            },
        )

        # check that our remote endpoint was passed custom headers, but not internal ones
        assert endpoint.called and endpoint.call_count == 1
        assert "x-harp-foo" not in endpoint.calls[0].request.headers
        assert endpoint.calls[0].request.headers["accept"] == "application/json"
        assert endpoint.calls[0].request.headers["vary"] == "custom"

        # check that the transaction was tagged with the expected values
        assert transaction_started_handler.called and transaction_started_handler.call_count == 1
        assert transaction_started_handler.call_args.args[0].transaction.tags == {"foo": "bar"}

        # check we got a valid response
        assert response.status == 200
        assert response.headers == {}
        assert response.body == b"Hello."

    async def test_get_next_url_for(self):
        controller = self.create_controller("http://example.com/base/")
        request = await _create_request(path="/foo/bar/")
        context = Mock()
        context.request = request
        base_url, full_url = await controller._get_next_url_for(context)
        assert base_url == "http://example.com/base/"
        assert full_url == "http://example.com/base/foo/bar/"


class TestHttpProxyControllerWithStorage(
    HttpProxyControllerTestFixtureMixin,
    StorageTestFixtureMixin,
    DispatcherTestFixtureMixin,
):
    async def call_controller(
        self,
        controller=None,
        /,
        *,
        dispatcher=None,
        engine=None,
        sql_storage=None,
        blob_storage=None,
        body=None,
        method="GET",
        headers=None,
    ) -> tuple[HttpRequest, HttpResponse]:
        dispatcher: IAsyncEventDispatcher = cast(IAsyncEventDispatcher, dispatcher or AsyncEventDispatcher())
        controller = controller or self.create_controller(dispatcher=dispatcher)
        worker = self.create_worker(dispatcher, engine, sql_storage, blob_storage)
        try:
            return await super().call_controller(controller, body=body, method=method, headers=headers)
        finally:
            await worker.wait_until_empty()

    async def _find_one_transaction_with_messages_from_storage(self, storage):
        # get transaction, request and response
        transactions = await storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1
        transaction = transactions[0]
        return transaction, transaction.messages[0], transaction.messages[1]

    @respx.mock
    async def test_basic_get(self, sql_storage: SqlStorage, blob_storage: IBlobStorage):
        self.mock_http_endpoint("http://example.com/", content="Hello.")

        # register the storage
        await self.call_controller(
            engine=sql_storage.engine,
            sql_storage=sql_storage,
            blob_storage=blob_storage,
        )

        transaction, request, response = await self._find_one_transaction_with_messages_from_storage(sql_storage)
        assert asdict(transaction) == {
            "id": ANY,
            "type": "http",
            "endpoint": None,
            "elapsed": ANY,
            "tpdex": ANY,
            "started_at": ANY,
            "finished_at": ANY,
            "messages": ANY,
            "tags": {},
            "extras": {
                "flags": [],
                "method": "GET",
                "status_class": "2xx",
                "cached": False,
                "no_cache": False,
            },
        }

        # request
        request_headers = await blob_storage.get(request.headers)
        assert request_headers.data == b""
        assert request_headers.content_type == "http/headers"
        assert (await blob_storage.get(request.body)).data == b""
        assert asdict(request) == {
            "id": 1,
            "transaction_id": ANY,
            "kind": "request",
            "summary": "GET / HTTP/1.1",
            "headers": "916ef336ce8ac9a91de41ce88c4b4bfc747b3ac9",
            "body": "adc83b19e793491b1c6ea0fd8b46cd9f32e592fc",
            "created_at": ANY,
        }

        # response
        response_headers = await blob_storage.get(response.headers)
        assert response_headers.data == b""
        assert response_headers.content_type == "http/headers"
        assert (await blob_storage.get(response.body)).data == b"Hello."
        assert asdict(response) == {
            "id": 2,
            "transaction_id": ANY,
            "kind": "response",
            "summary": "HTTP/1.1 200 OK",
            "headers": "916ef336ce8ac9a91de41ce88c4b4bfc747b3ac9",
            "body": "6ffdd89703735cc316470566467b816446f008ce",
            "created_at": ANY,
        }

    @respx.mock
    async def test_get_with_tags(self, sql_storage: SqlStorage, blob_storage: IBlobStorage):
        self.mock_http_endpoint("http://example.com/", content="Hello.")

        # call our controller
        await self.call_controller(
            engine=sql_storage.engine,
            sql_storage=sql_storage,
            blob_storage=blob_storage,
            headers={
                "x-harp-foo": "bar",
                "accept": "application/json",
                "vary": "custom",
            },
        )

        transaction, request, response = await self._find_one_transaction_with_messages_from_storage(sql_storage)

        assert asdict(transaction) == {
            "id": ANY,
            "type": "http",
            "endpoint": None,
            "elapsed": ANY,
            "tpdex": ANY,
            "started_at": ANY,
            "finished_at": ANY,
            "messages": ANY,
            "tags": {"foo": "bar"},
            "extras": {
                "flags": [],
                "method": "GET",
                "status_class": "2xx",
                "cached": False,
                "no_cache": False,
            },
        }

        request_headers = await blob_storage.get(request.headers)
        assert request_headers.data == (b"accept: application/json\nvary: custom")
        assert request_headers.content_type == "http/headers"
        assert (await blob_storage.get(request.body)).data == b""
        assert asdict(request) == {
            "id": 1,
            "transaction_id": ANY,
            "kind": "request",
            "summary": "GET / HTTP/1.1",
            "headers": "62ccbc3696048078cc4ced90d9239c1d4abc9e49",
            "body": "adc83b19e793491b1c6ea0fd8b46cd9f32e592fc",
            "created_at": ANY,
        }

        response_headers = await blob_storage.get(response.headers)
        assert response_headers.data == b""
        assert response_headers.content_type == "http/headers"
        assert (await blob_storage.get(response.body)).data == b"Hello."
        assert asdict(response) == {
            "id": 2,
            "transaction_id": ANY,
            "kind": "response",
            "summary": "HTTP/1.1 200 OK",
            "headers": "916ef336ce8ac9a91de41ce88c4b4bfc747b3ac9",
            "body": "6ffdd89703735cc316470566467b816446f008ce",
            "created_at": ANY,
        }



from .blob_storage import IBlobStorage
from .helpers import TransactionsGroupedByTimeBucket
from .storage import IStorage

__all__ = [
    "IBlobStorage",
    "IStorage",
    "TransactionsGroupedByTimeBucket",
]



from ssl import SSLCertVerificationError

import httpcore
import httpx

BREAK_ON_HTTP_4xx = "http_4xx"
BREAK_ON_HTTP_5xx = "http_5xx"
BREAK_ON_NETWORK_ERROR = "network_error"
BREAK_ON_UNHANDLED_EXCEPTION = "unhandled_exception"

DOWN = -1
CHECKING = 0
UP = 1

DEFAULT_POOL = "default"
FALLBACK_POOL = "fallback"
AVAILABLE_POOLS = {DEFAULT_POOL, FALLBACK_POOL}

ALL_BREAK_ON_VALUES = {
    BREAK_ON_HTTP_4xx,
    BREAK_ON_HTTP_5xx,
    BREAK_ON_NETWORK_ERROR,
    BREAK_ON_UNHANDLED_EXCEPTION,
}

ERR_UNHANDLED_STATUS_CODE = 500
ERR_UNHANDLED_MESSAGE = "Unhandled error"
ERR_UNHANDLED_VERBOSE_MESSAGE = "Internal Server Error (unhandled error)"

ERR_UNAVAILABLE_STATUS_CODE = 503
ERR_UNAVAILABLE_MESSAGE = "Unavailable"
ERR_UNAVAILABLE_VERBOSE = "Service Unavailable (remote server unavailable)"

ERR_TIMEOUT_STATUS_CODE = 504
ERR_TIMEOUT_MESSAGE = "Timeout"
ERR_TIMEOUT_VERBOSE_MESSAGE = "Gateway Timeout (remote server timeout)"

ERR_REMOTE_STATUS_CODE = 502
ERR_REMOTE_MESSAGE = "Remote server disconnected"
ERR_REMOTE_VERBOSE_MESSAGE = "Bad Gateway (remote server disconnected)"

ERR_SSL_STATUS_CODE = 526
ERR_SSL_MESSAGE = "Invalid SSL certificate"
ERR_SSL_VERBOSE_MESSAGE = "Invalid SSL Certificate (certificate verification failed)"

NETWORK_ERRORS = {
    httpcore.NetworkError: (
        ERR_UNAVAILABLE_STATUS_CODE,
        ERR_UNAVAILABLE_MESSAGE,
        ERR_UNAVAILABLE_VERBOSE,
    ),
    httpx.NetworkError: (
        ERR_UNAVAILABLE_STATUS_CODE,
        ERR_UNAVAILABLE_MESSAGE,
        ERR_UNAVAILABLE_VERBOSE,
    ),
    httpcore.TimeoutException: (
        ERR_TIMEOUT_STATUS_CODE,
        ERR_TIMEOUT_MESSAGE,
        ERR_TIMEOUT_VERBOSE_MESSAGE,
    ),
    httpx.TimeoutException: (
        ERR_TIMEOUT_STATUS_CODE,
        ERR_TIMEOUT_MESSAGE,
        ERR_TIMEOUT_VERBOSE_MESSAGE,
    ),
    httpcore.RemoteProtocolError: (
        ERR_REMOTE_STATUS_CODE,
        ERR_REMOTE_MESSAGE,
        ERR_REMOTE_VERBOSE_MESSAGE,
    ),
    httpx.RemoteProtocolError: (
        ERR_REMOTE_STATUS_CODE,
        ERR_REMOTE_MESSAGE,
        ERR_REMOTE_VERBOSE_MESSAGE,
    ),
    SSLCertVerificationError: (
        ERR_SSL_STATUS_CODE,
        ERR_SSL_MESSAGE,
        ERR_SSL_VERBOSE_MESSAGE,
    ),
}



import zoneinfo
from datetime import datetime
from unittest.mock import ANY, Mock

import freezegun
import orjson
import pytest
from multidict import MultiDict

from harp.http import HttpRequest
from harp.utils.testing.communicators import ASGICommunicator
from harp.utils.testing.mixins import ControllerThroughASGIFixtureMixin
from harp_apps.dashboard.controllers.overview import OverviewController, time_bucket_for_range
from harp_apps.storage.services.sql import SqlStorage
from harp_apps.storage.utils.testing.mixins import StorageTestFixtureMixin


class OverviewControllerTestFixtureMixin:
    @pytest.fixture
    def controller(self, sql_storage):
        return OverviewController(storage=sql_storage, handle_errors=False)


class _Any24IntegerValues(object):
    "A helper object that compares equal to everything."

    def __eq__(self, other):
        return len(other) == 24 and all(lambda x: isinstance(x.get("value"), int) for x in other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "<Any24Values>"


Any24IntegerValues = _Any24IntegerValues()

parametrize_with_now = pytest.mark.parametrize(
    "now",
    [
        # datetime(2023, 6, 21, 12, 10, 0),
        # datetime(2023, 6, 21, 12, 0, 0),
        datetime(2023, 6, 21, 12, 0, 0, tzinfo=zoneinfo.ZoneInfo("Australia/Sydney")),
    ],
)

parametrize_with_time_range = pytest.mark.parametrize("time_range", list(time_bucket_for_range.keys()))
parametrize_with_time_range = pytest.mark.parametrize("time_range", ["7d"])


class TestOverviewController(OverviewControllerTestFixtureMixin, StorageTestFixtureMixin):
    @parametrize_with_now
    async def test_get_summary_data_no_transactions(self, controller: OverviewController, now):
        request = Mock(spec=HttpRequest, query=MultiDict())
        with freezegun.freeze_time(now):
            response = await controller.get_summary_data(request)
        assert response == {
            "tpdex": {"data": Any24IntegerValues, "mean": 100},
            "errors": {"data": Any24IntegerValues, "period": "day", "rate": 0},
            "transactions": {"data": Any24IntegerValues, "period": "day", "rate": 0},
        }

    @parametrize_with_now
    async def test_get_summary_data_with_a_few_transactions(
        self, controller: OverviewController, sql_storage: SqlStorage, now
    ):
        with freezegun.freeze_time(now):
            await self.create_transaction(sql_storage)
            await self.create_transaction(sql_storage)
            await self.create_transaction(sql_storage)
        request = Mock(spec=HttpRequest, query=MultiDict())
        with freezegun.freeze_time(now):
            response = await controller.get_summary_data(request)
        assert response == {
            "tpdex": {"data": Any24IntegerValues, "mean": 100},
            "errors": {"data": Any24IntegerValues, "period": "day", "rate": 0},
            "transactions": {"data": Any24IntegerValues, "period": "day", "rate": 3},
        }

    @parametrize_with_now
    @parametrize_with_time_range
    async def test_get_overview_data_with_a_few_transactions(
        self, controller: OverviewController, sql_storage: SqlStorage, now, time_range
    ):
        with freezegun.freeze_time(now):
            await self.create_transaction(sql_storage)
            await self.create_transaction(sql_storage)
            await self.create_transaction(sql_storage)
        request = Mock(spec=HttpRequest, query=MultiDict(endpoint="/", timeRange=time_range))
        with freezegun.freeze_time(now):
            response = await controller.get_overview_data(request)
        assert response == {
            "count": 3,
            "errors": {"count": 0, "rate": 0.0},
            "meanDuration": 0.0,
            "meanTpdex": 100,
            "timeRange": "7d",
            "transactions": ANY,
        }
        assert response["transactions"][-1]["count"] == 3


class TestOverviewControllerThroughASGI(
    OverviewControllerTestFixtureMixin,
    StorageTestFixtureMixin,
    ControllerThroughASGIFixtureMixin,
):
    @parametrize_with_now
    async def test_get_summary_data_with_a_few_transactions_using_asgi(
        self, client: ASGICommunicator, sql_storage: SqlStorage, now
    ):
        with freezegun.freeze_time(now):
            await self.create_transaction(sql_storage)
            await self.create_transaction(sql_storage)
            await self.create_transaction(sql_storage)

            response = await client.http_get("/api/overview/summary")

        assert response["status"] == 200
        assert response["headers"] == ((b"content-type", b"application/json"),)

        assert orjson.loads(response["body"]) == {
            "tpdex": {"data": Any24IntegerValues, "mean": 100},
            "errors": {"data": Any24IntegerValues, "period": "day", "rate": "0"},
            "transactions": {"data": Any24IntegerValues, "period": "day", "rate": "3"},
        }



import dataclasses
from datetime import date

from dataclasses_jsonschema import JsonSchemaMixin


@dataclasses.dataclass(kw_only=True)
class TransactionsByDate(JsonSchemaMixin):
    date: date
    transactions: int
    errors: int



from functools import cached_property
from urllib.parse import urlencode

from httpx import codes

from .errors import HttpError
from .requests import HttpRequest
from .responses import HttpResponse
from .typing import BaseHttpMessage, BaseMessage, MessageSerializer


class BaseHttpMessageSerializer(MessageSerializer):
    """
    Base serializer for HTTP messages.

    """

    def __init__(self, message: BaseHttpMessage):
        self.wrapped = message

    @cached_property
    def summary(self) -> str:
        raise NotImplementedError

    @cached_property
    def headers(self) -> str:
        return "\n".join([f"{k}: {v}" for k, v in self.wrapped.headers.items()])

    @cached_property
    def body(self) -> bytes:
        return self.wrapped.body


class HttpRequestSerializer(BaseHttpMessageSerializer):
    """
    Serialize an HTTP request object into string representations for different message parts:

    - summary: the first line of the request message (e.g. "GET / HTTP/1.1")
    - headers: the headers of the request message (e.g. "Host: localhost:4080\nConnection: keep-alive\n...")
    - body: the body of the request message (e.g. b'{"foo": "bar"}')

    The main goal of this serializer is to prepare a request message for storage.

    """

    wrapped: HttpRequest

    @cached_property
    def query_string(self) -> str:
        return urlencode(self.wrapped.query) if self.wrapped.query else ""

    @cached_property
    def summary(self) -> str:
        query = f"?{self.query_string}" if self.query_string else ""
        return f"{self.wrapped.method} {self.wrapped.path}{query} HTTP/1.1"


class HttpResponseSerializer(BaseHttpMessageSerializer):
    """
    Serialize an HTTP response object into string representations for different message parts:

    - summary: the first line of the response message (e.g. "HTTP/1.1 200 OK")
    - headers: the headers of the response message (e.g. "Content-Type: text/plain\nContent-Length: 13\n...")
    - body: the body of the response message (e.g. b'Hello, world!')

    The main goal of this serializer is to prepare a response message for storage.

    """

    wrapped: HttpResponse

    @cached_property
    def summary(self) -> str:
        reason = codes.get_reason_phrase(self.wrapped.status)
        return f"HTTP/1.1 {self.wrapped.status} {reason}"


class HttpErrorSerializer(BaseHttpMessageSerializer):
    """
    Serialize an HTTP error object into string representations for different message parts:

    - summary: the error message
    - headers: empty
    - body: stack trace (xxx this may change, maybe too much info and too much internal)

    The main goal of this serializer is to prepare an error message for storage.

    """

    wrapped: HttpError

    @cached_property
    def summary(self) -> str:
        return self.wrapped.message


def get_serializer_for(message: BaseMessage) -> MessageSerializer:
    if isinstance(message, HttpRequest):
        return HttpRequestSerializer(message)

    if isinstance(message, HttpResponse):
        return HttpResponseSerializer(message)

    if isinstance(message, HttpError):
        return HttpErrorSerializer(message)

    raise ValueError(f"No serializer available for message type: {type(message)}")



"""messages indexes

Revision ID: d4cd37d29d7a
Revises: 0b4d9cb71c38
Create Date: 2024-08-09 06:14:14.691024

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = "d4cd37d29d7a"
down_revision: Union[str, None] = "0b4d9cb71c38"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.create_unique_constraint(None, "blobs", ["id"])
    op.alter_column("messages", "kind", existing_type=sa.VARCHAR(length=10), nullable=False)
    op.alter_column("messages", "summary", existing_type=sa.VARCHAR(length=255), nullable=False)
    op.alter_column("messages", "headers", existing_type=sa.VARCHAR(length=40), nullable=False)
    op.alter_column(
        "messages",
        "created_at",
        existing_type=postgresql.TIMESTAMP(timezone=True),
        nullable=False,
    )
    op.alter_column(
        "messages",
        "transaction_id",
        existing_type=sa.VARCHAR(length=27),
        nullable=False,
    )
    op.create_index("ix_transaction_id", "messages", ["transaction_id"], unique=False)
    op.create_unique_constraint(None, "messages", ["id"])
    op.create_unique_constraint(None, "metrics", ["id"])
    op.create_unique_constraint("__values_uc", "tag_values", ["tag_id", "value"])
    op.create_unique_constraint(None, "tag_values", ["id"])
    op.create_unique_constraint(None, "tags", ["id"])
    op.create_unique_constraint(None, "trans_user_flags", ["id"])
    op.alter_column("transactions", "type", existing_type=sa.VARCHAR(length=10), nullable=False)
    op.alter_column(
        "transactions",
        "started_at",
        existing_type=postgresql.TIMESTAMP(timezone=True),
        nullable=False,
    )
    op.create_unique_constraint(None, "transactions", ["id"])
    op.create_unique_constraint(None, "users", ["id"])


def downgrade() -> None:
    op.drop_constraint(None, "users", type_="unique")
    op.drop_constraint(None, "transactions", type_="unique")
    op.alter_column(
        "transactions",
        "started_at",
        existing_type=postgresql.TIMESTAMP(timezone=True),
        nullable=True,
    )
    op.alter_column("transactions", "type", existing_type=sa.VARCHAR(length=10), nullable=True)
    op.drop_constraint(None, "trans_user_flags", type_="unique")
    op.drop_constraint(None, "tags", type_="unique")
    op.drop_constraint(None, "tag_values", type_="unique")
    op.drop_constraint("__values_uc", "tag_values", type_="unique")
    op.drop_constraint(None, "metrics", type_="unique")
    op.drop_constraint(None, "messages", type_="unique")
    op.drop_index("ix_transaction_id", table_name="messages")
    op.alter_column("messages", "transaction_id", existing_type=sa.VARCHAR(length=27), nullable=True)
    op.alter_column(
        "messages",
        "created_at",
        existing_type=postgresql.TIMESTAMP(timezone=True),
        nullable=True,
    )
    op.alter_column("messages", "headers", existing_type=sa.VARCHAR(length=40), nullable=True)
    op.alter_column("messages", "summary", existing_type=sa.VARCHAR(length=255), nullable=True)
    op.alter_column("messages", "kind", existing_type=sa.VARCHAR(length=10), nullable=True)
    op.drop_constraint(None, "blobs", type_="unique")



from typing import TYPE_CHECKING, Callable, Type, cast

from asgiref.typing import ASGIApplication
from whistle import IAsyncEventDispatcher

from harp import __revision__, __version__, get_logger
from harp.asgi import ASGIKernel
from harp.asgi.events import EVENT_CORE_REQUEST, EVENT_CORE_VIEW
from harp.event_dispatcher import LoggingAsyncEventDispatcher
from harp.services import Container, Services
from harp.typing import GlobalSettings
from harp.utils.network import Bind
from harp.views.json import on_json_response

from .. import ApplicationsRegistry
from ..events import (
    EVENT_BIND,
    EVENT_BOUND,
    EVENT_READY,
    EVENT_SHUTDOWN,
    OnBindEvent,
    OnBoundEvent,
    OnReadyEvent,
    OnShutdownEvent,
)

if TYPE_CHECKING:
    from harp.controllers import ProxyControllerResolver

logger = get_logger(__name__)


class System:
    """
    The core, global-like system objects, encapsulated together. Useful to manipulate the system from a high-level
    perspective, either from a server adapter point of view or from tests.

    Bootstrapping will create a System instance, which will be used to start the various subsystems. Then, the
    subsystems will work on their own, and should not need to have a system view.

    Attributes:
        config (GlobalSettings): The global configuration settings for the application.
        dispatcher (IAsyncEventDispatcher): The event dispatcher for handling application-wide events.
        provider (Services): The service provider for dependency injection and service management.
        kernel (ASGIKernel): The ASGI kernel for handling HTTP requests.
        binds (list[Bind]): A list of network binds specifying where the application should listen for incoming requests.

    Methods:
        dispose(): Asynchronously disposes of the system resources, primarily the ASGI kernel.
    """

    def __init__(
        self,
        config: GlobalSettings,
        dispatcher: IAsyncEventDispatcher,
        provider: Services,
        /,
        *,
        asgi_app: ASGIApplication,
        binds: list[Bind],
    ):
        self._config = config
        self._dispatcher = dispatcher
        self._provider = provider
        self._asgi_app = asgi_app
        self._binds = binds

    @property
    def config(self) -> GlobalSettings:
        return self._config

    @property
    def dispatcher(self) -> IAsyncEventDispatcher:
        return self._dispatcher

    @property
    def provider(self) -> Services:
        return self._provider

    @property
    def asgi_app(self) -> ASGIApplication:
        return self._asgi_app

    @property
    def binds(self) -> list[Bind]:
        return self._binds

    async def dispose(self):
        """
        Asynchronously disposes of the system resources, primarily the ASGI kernel.
        """
        if self.asgi_app:
            try:
                event = OnShutdownEvent(self.asgi_app, self.provider)
                await self.dispatcher.adispatch(EVENT_SHUTDOWN, event)
            except Exception as exc:
                logger.fatal("💣 Fatal while dispatching «%s» event: %s", EVENT_SHUTDOWN, exc)
                raise
            finally:
                self._asgi_app = None


class SystemBuilder:
    """
    A builder class for constructing the System instance, which represents the core of the HARP application.

    This class is responsible for assembling the necessary components of the System, including the global configuration,
    event dispatcher, service provider, and ASGI kernel, based on the provided configuration.

    Attributes:
        AsyncEventDispatcherType (Type[IAsyncEventDispatcher]): The type of the event dispatcher to use.
        ContainerType (Type[Container]): The type of the container to use for dependency injection.
        KernelType (Type[ASGIKernel]): The type of the ASGI kernel to use for handling HTTP requests.
        configuration_builder (ConfigurationBuilder): The configuration builder used to assemble the global configuration.
        hostname (str): The hostname where the application should listen for incoming requests.

    Methods:
        abuild() -> System: Asynchronously builds and returns an instance of the System class.
    """

    AsyncEventDispatcherType: Type[IAsyncEventDispatcher] = LoggingAsyncEventDispatcher
    ContainerType: Type[Container] = Container
    KernelType: Type[ASGIKernel] = ASGIKernel

    def __init__(
        self,
        applications=ApplicationsRegistry | Callable[[], ApplicationsRegistry],
        configuration=GlobalSettings | Callable[[], GlobalSettings],
        /,
        *,
        hostname: str = "[::]",
    ):
        self._applications = applications
        self._configuration = configuration
        self.hostname = hostname

    @property
    def applications(self) -> ApplicationsRegistry:
        if callable(self._applications):
            self._applications = self._applications()
        return self._applications

    @property
    def configuration(self) -> GlobalSettings:
        if callable(self._configuration):
            self._configuration = self._configuration()
        return self._configuration

    async def abuild(self) -> System:
        """
        Asynchronously builds and returns an instance of the System class, ready for use.

        Returns:
            System: An instance of the System class, assembled based on the provided configuration and components.
        """
        logger.info(f"🎙  HARP v.{__version__} ({__revision__})")

        # Get lazy configuration.
        config = self.configuration
        logger.info(f"📦 {', '.join(self.applications.keys())}")
        for name, app in self.applications.items():
            logger.debug(f'... "{name}" application loaded from "{app.path}"')

        # Prepare and dispatch «bind» event.
        dispatcher = self.build_dispatcher()
        container = self.build_container(config, dispatcher)
        await self.dispatch_bind_event(dispatcher, container, config)

        # todo: this looks like not the right place, should not be tightly coupled to the factory
        from harp.controllers import ProxyControllerResolver

        controller_resolver = ProxyControllerResolver()
        container.add_instance(controller_resolver, ProxyControllerResolver)

        # Prepare and dispatch «bound» event.
        provider = container.build_provider()
        await self.dispatch_bound_event(dispatcher, provider, controller_resolver)

        # Build the kernel and dispatch «ready» event.
        event = await self.dispatch_ready_event(dispatcher, provider, controller_resolver)

        # Send back a coherent view of the system.
        return System(config, dispatcher, provider, asgi_app=event.asgi_app, binds=event.binds)

    def build_dispatcher(self):
        dispatcher = cast(IAsyncEventDispatcher, self.AsyncEventDispatcherType())
        self.applications.register_events(dispatcher)

        self.__setup_core_events(dispatcher)

        return dispatcher

    def __setup_core_events(self, dispatcher):
        # todo move into core or extension, this is not system or proxy related
        from harp.controllers.default import on_health_request

        dispatcher.add_listener(EVENT_CORE_REQUEST, on_health_request, priority=-100)
        dispatcher.add_listener(EVENT_CORE_VIEW, on_json_response)

    def build_container(self, config, dispatcher):
        container = cast(Container, self.ContainerType())
        container.add_instance(dispatcher, IAsyncEventDispatcher)
        container.add_instance(config, GlobalSettings)
        self.applications.register_services(container, config)
        return container

    async def dispatch_bind_event(
        self,
        dispatcher: IAsyncEventDispatcher,
        container: Container,
        config,
    ) -> OnBindEvent:
        # dispatch "bind" event: this is the last chance to add services to the container
        try:
            return cast(
                OnBindEvent,
                await dispatcher.adispatch(EVENT_BIND, OnBindEvent(container, config)),
            )
        except Exception as exc:
            logger.fatal("💣 Fatal while dispatching «%s» event: %s", EVENT_BIND, exc)
            raise

    async def dispatch_bound_event(
        self,
        dispatcher: IAsyncEventDispatcher,
        provider: Services,
        resolver: "ProxyControllerResolver",
    ) -> OnBoundEvent:
        # dispatch "bound" event: you get a resolved container, do your thing
        try:
            return cast(
                OnBoundEvent,
                await dispatcher.adispatch(EVENT_BOUND, OnBoundEvent(provider, resolver)),
            )
        except Exception as exc:
            logger.fatal("💣 Fatal while dispatching «%s» event: %s", EVENT_BOUND, exc)
            raise

    async def dispatch_ready_event(
        self,
        dispatcher: IAsyncEventDispatcher,
        provider: Services,
        controller_resolver: "ProxyControllerResolver",
    ):
        # todo: this should be instanciated by the service container, probably using a factory to dispatch the event,
        #  etc. Binds should not sit here, but where?
        asgi_app = self.KernelType(dispatcher=dispatcher, resolver=controller_resolver)
        binds = [Bind(host=self.hostname, port=port) for port in controller_resolver.ports]
        try:
            return cast(
                OnReadyEvent,
                await dispatcher.adispatch(EVENT_READY, OnReadyEvent(provider, asgi_app, binds)),
            )
        except Exception as exc:
            logger.fatal("💣 Fatal while dispatching «%s» event: %s", EVENT_READY, exc)
            raise



from httpx import ByteStream

from harp.http import HttpResponse


class BaseHttpResponseTest:
    """
    Base class for testing HTTP responses.
    """

    ResponseType = HttpResponse

    def create_response(self, body, **kwargs) -> HttpResponse:
        return self.ResponseType(body, **kwargs)


class TestHttpResponseStatus(BaseHttpResponseTest):
    """
    Test status-related stuff on HTTP responses.
    """

    def test_default(self):
        response = self.create_response(body=b"")
        assert response.status == 200

    def test_custom_status(self):
        response = self.create_response(body=b"", status=404)
        assert response.status == 404


class TestHttpResponseHeaders(BaseHttpResponseTest):
    """
    Test headers-related stuff on HTTP responses.
    """

    def test_headers(self):
        response = self.create_response(
            body=b"",
            headers={
                "content-type": "application/json",
                "x-custom-header": "value",
            },
        )

        assert response.headers == {
            "content-type": "application/json",
            "x-custom-header": "value",
        }

    def test_content_type(self):
        response = self.create_response(body=b"", content_type="application/json")
        assert response.content_type == "application/json"


class TestHttpResponseBody(BaseHttpResponseTest):
    async def test_body_empty(self):
        response = self.create_response(body=b"")
        await response.aread()
        assert response.body == b""

    async def test_body_one_chunk(self):
        response = self.create_response(body=b"foobar")
        await response.aread()
        assert response.body == b"foobar"

    async def test_body_can_be_read_more_than_once(self):
        response = self.create_response(body=b"foobar")
        await response.aread()
        await response.aread()
        await response.aread()
        assert response.body == b"foobar"

    async def test_stream_can_be_accessed_before_reading_body(self):
        response = self.create_response(body=b"foobarbaz")
        assert [chunk async for chunk in response.stream] == [b"foobarbaz"]
        response.stream = ByteStream(b"foobarbaz")
        await response.aread()
        assert response.body == b"foobarbaz"

    async def test_stream_can_be_accessed_after_reading_body(self):
        response = self.create_response(body=b"foobarbaz")
        await response.aread()
        assert [chunk async for chunk in response.stream] == [b"foobarbaz"]
        assert response.body == b"foobarbaz"
        assert isinstance(response.body, bytes)
        assert isinstance(response.stream, ByteStream)


class TestHttpResponseReasonPhrase(BaseHttpResponseTest):
    """
    Test reason_phrase property on HTTP responses.
    """

    def test_reason_phrase_default_from_status(self):
        """When no reason_phrase extension is set, it should use httpx's default for the status code."""
        response = self.create_response(body=b"", status=200)
        assert response.reason_phrase == "OK"

    def test_reason_phrase_custom_status(self):
        """Different status codes should have different default reason phrases."""
        response = self.create_response(body=b"", status=404)
        assert response.reason_phrase == "Not Found"

        response = self.create_response(body=b"", status=502)
        assert response.reason_phrase == "Bad Gateway"

    def test_reason_phrase_from_bytes_extension(self):
        """When reason_phrase extension is set as bytes, it should decode it."""
        response = self.create_response(body=b"", status=502, extensions={"reason_phrase": b"Custom Gateway Error"})
        assert response.reason_phrase == "Custom Gateway Error"

    def test_reason_phrase_from_str_extension(self):
        """When reason_phrase extension is set as str, it should return it as-is."""
        response = self.create_response(body=b"", status=502, extensions={"reason_phrase": "Custom Gateway Error"})
        assert response.reason_phrase == "Custom Gateway Error"



from harp.http import HttpRequest
from harp_apps.proxy.constants import CHECKING, DOWN, UP


def humanize_status(status):
    return {CHECKING: "checking", UP: "up", DOWN: "down"}.get(status, "unknown")


def extract_tags_from_request(request: HttpRequest):
    """
    Convert special request headers (x-harp-*) into tags (key-value pairs) that we'll attach to the
    transaction. Headers are "consumed", meaning they are removed from the request headers.
    """

    tags = {}
    headers_to_remove = []

    for header in request.headers:
        lower_header = header.lower()
        if lower_header.startswith("x-harp-"):
            tags[lower_header[7:]] = request.headers[header]
            headers_to_remove.append(header)

    for header in headers_to_remove:
        request.headers.pop(header)

    return tags



import asyncio
import os
from concurrent.futures import ThreadPoolExecutor
from typing import Union

from pydantic_core import MultiHostUrl, Url
from sqlalchemy import URL, make_url, text
from sqlalchemy.exc import OperationalError

from harp import get_logger
from harp.config import ConfigurationBuilder
from harp_apps import storage
from harp_apps.storage.models import Base, Message, Transaction

logger = get_logger(__name__)


def create_alembic_config(url: Union[str, URL, Url, MultiHostUrl]):
    """Create our alembic configuration object, to run alembic commands."""
    from alembic.config import Config as AlembicConfig

    if isinstance(url, (Url, MultiHostUrl)):
        url = str(url)

    url = make_url(url)

    alembic_cfg = AlembicConfig()
    alembic_cfg.set_main_option("script_location", os.path.join(storage.__path__[0], "migrations"))
    alembic_cfg.set_main_option(
        "file_template",
        "%%(year)d%%(month).2d%%(day).2d%%(hour).2d%%(minute).2d%%(second).2d_%%(rev)s_%%(slug)s",
    )
    alembic_cfg.set_main_option("truncate_slug_length", "40")
    alembic_cfg.set_main_option("output_encoding", "utf-8")
    alembic_cfg.set_main_option("sqlalchemy.url", url.render_as_string(hide_password=False))
    alembic_cfg.set_main_option("configure_logger", "false")

    return alembic_cfg


def create_harp_settings_with_storage_from_command_line_options(kwargs):
    """Create a Harp configuration object using common server command line options (--set...) with the sqlalchemy
    storage application installed so that the storage is properly configured."""

    from harp.commandline.options.server import CommonServerOptions

    options = CommonServerOptions(**kwargs)

    builder = ConfigurationBuilder.from_commandline_options(options)
    builder.applications.add("storage")

    return builder.build()


async def do_reset(engine):
    logger.info("🛢 [db:reset] dropping all tables.")
    async with engine.begin() as conn:
        await conn.run_sync(Base.metadata.drop_all)
        await conn.execute(text("DROP TABLE IF EXISTS alembic_version;"))


async def _do_migrate(engine, *, migrator, reset=False):
    logger.info(f"🛢 Starting database migrations... (dialect={engine.dialect.name}, reset={reset}).")
    if reset:
        await do_reset(engine)

    if engine.dialect.name == "sqlite":
        logger.debug("🛢 [db:migrate dialect=sqlite] creating all tables (without alembic).")
        async with engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)
            await conn.commit()

    elif migrator:
        # alembic manages migrations except for sqlite, because it's not trivial to make them work and an env using
        # sqlite does not really need to support upgrades (drop/recreate is fine when harp is upgraded).
        logger.info("🛢 [db:migrate] Running database migrations...")

        with ThreadPoolExecutor() as executor:
            await asyncio.get_event_loop().run_in_executor(executor, migrator)

    if engine.dialect.name == "mysql":
        logger.debug("🛢 [db:migrate dialect=mysql] creating fulltext indexes.")

        try:
            async with engine.begin() as conn:
                await conn.execute(
                    text(f"CREATE FULLTEXT INDEX endpoint_ft_index ON {Transaction.__tablename__} (endpoint);")
                )
                # Create the full text index for messages.summary
                await conn.execute(
                    text(f"CREATE FULLTEXT INDEX summary_ft_index ON {Message.__tablename__} (summary);")
                )
                await conn.commit()
        except OperationalError as e:
            # check for duplicate key error
            if e.orig and e.orig.args[0] == 1061:
                pass
            else:
                raise e

    logger.debug("🛢 [db:migrate] Done.")


async def do_migrate(engine, *, migrator, reset=False):
    try:
        await _do_migrate(engine, migrator=migrator, reset=reset)
    except Exception as e:
        logger.error(f"🛢 [db:migrate] Migrations failed: {e}")
        raise RuntimeError(f"Could not run migrations ({e}).") from e



import shlex
import subprocess
import time
from typing import List

from harp import get_logger

logger = get_logger(__name__)


def start_subprocess_with_retry(
    command: List[str], max_retries: int = 3, retry_delay: float = 1.0, startup_check_delay: float = 0.5
) -> subprocess.Popen:
    """Start a subprocess with retry logic.

    Args:
        command: Command to execute.
        max_retries: Maximum number of retry attempts.
        retry_delay: Delay between retries in seconds.
        startup_check_delay: Delay before checking if process started successfully.

    Returns:
        The started subprocess.

    Raises:
        RuntimeError: If subprocess fails to start after max retries.
    """
    for attempt in range(max_retries):
        try:
            logger.info(f"Starting subprocess (attempt {attempt + 1}/{max_retries}): {shlex.join(command)}")
            process = subprocess.Popen(command)

            # Give the process time to start and check if it's still running
            time.sleep(startup_check_delay)
            poll_result = process.poll()
            if poll_result is None:
                logger.info("Subprocess started successfully")
                return process
            else:
                logger.warning(f"Subprocess exited with code {poll_result}")

        except Exception as e:
            logger.error(f"Failed to start subprocess: {e}")

        if attempt < max_retries - 1:
            logger.info(f"Retrying in {retry_delay} seconds...")
            time.sleep(retry_delay)

    raise RuntimeError(f"Failed to start subprocess after {max_retries} attempts")



"""messages summary from varchar to text

Revision ID: 4ab341a710f4
Revises: 0b4d9cb71c38
Create Date: 2024-09-06 09:56:42.404282

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "4ab341a710f4"
down_revision: Union[str, None] = "0b4d9cb71c38"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "messages", "summary", existing_type=sa.VARCHAR(length=255), type_=sa.Text(), existing_nullable=True
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "messages", "summary", existing_type=sa.Text(), type_=sa.VARCHAR(length=255), existing_nullable=True
    )
    # ### end Alembic commands ###



import logging.config
import os
from typing import Any, Optional

import structlog

shared_processors = [
    structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    structlog.stdlib.ExtraAdder(),
    structlog.processors.TimeStamper(fmt="iso"),
]

processors = [
    structlog.stdlib.filter_by_level,
    structlog.contextvars.merge_contextvars,
    *shared_processors,
    structlog.dev.set_exc_info,
    structlog.stdlib.PositionalArgumentsFormatter(),
    structlog.processors.StackInfoRenderer(),
    structlog.processors.UnicodeDecoder(),
    structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
]

try:
    _exc_formatter_options = {"width": os.get_terminal_size()[0]}
except OSError:
    _exc_formatter_options = {}


LOGGING_FORMATTERS = {
    "json": {
        "()": structlog.stdlib.ProcessorFormatter,
        "processor": structlog.processors.JSONRenderer(),
        "foreign_pre_chain": shared_processors,
    },
    "plain": {
        "()": structlog.stdlib.ProcessorFormatter,
        "processor": structlog.dev.ConsoleRenderer(
            exception_formatter=structlog.dev.plain_traceback,
            colors=False,
        ),
        "foreign_pre_chain": shared_processors,
    },
    "pretty": {
        "()": structlog.stdlib.ProcessorFormatter,
        "processor": structlog.dev.ConsoleRenderer(
            exception_formatter=structlog.dev.plain_traceback,
        ),
        "foreign_pre_chain": shared_processors,
    },
    "keyvalue": {
        "()": structlog.stdlib.ProcessorFormatter,
        "processor": structlog.processors.KeyValueRenderer(key_order=["timestamp", "level", "event", "logger"]),
        "foreign_pre_chain": shared_processors,
    },
}

DEFAULT_LOGGING_FORMAT = "pretty"
LOGGING_FORMAT = os.environ.get("LOGGING_FORMAT", DEFAULT_LOGGING_FORMAT)
if LOGGING_FORMAT not in LOGGING_FORMATTERS:
    LOGGING_FORMAT = DEFAULT_LOGGING_FORMAT


def _get_logging_level(name: Optional[str], *, default="warning"):
    varname = "_".join(filter(None, ("LOGGING", name.upper() if name is not None else None)))
    return os.environ.get(varname, default).upper()


logging_config = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": LOGGING_FORMATTERS,
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": LOGGING_FORMAT,
        },
    },
    "root": {
        "handlers": ["console"],
        "level": _get_logging_level(None, default="info"),
    },
    "loggers": {
        "harp": {"level": _get_logging_level("harp", default="info")},
        "harp.event_dispatcher": {"level": _get_logging_level("events")},
        "harp_apps": {"level": _get_logging_level("harp")},
        "harp_apps.http_client": {"level": _get_logging_level("http_client")},
        "harp_apps.proxy": {"level": _get_logging_level("proxy")},
        "harp_apps.storage": {"level": _get_logging_level("storage")},
        "httpcore": {"level": _get_logging_level("http_core")},
        "httpx": {"level": _get_logging_level("http_core")},
        "hypercorn.access": {"level": _get_logging_level("http", default="info")},
        "hypercorn.error": {"level": _get_logging_level("http", default="info")},
        "sqlalchemy.engine": {"level": _get_logging_level("sql")},
    },
}


logging.config.dictConfig(logging_config)

structlog.configure(
    logger_factory=structlog.stdlib.LoggerFactory(),
    wrapper_class=structlog.stdlib.BoundLogger,
    processors=processors,
    cache_logger_on_first_use=True,
)


def get_logger(name, *args: Any, **initial_values: Any) -> Any:
    if name == "env_py":
        name = "__migrations__"
    else:
        try:
            pkg, mod = name.rsplit(".", 1)
        except ValueError:
            return structlog.get_logger(name, *args, **initial_values)
        if mod in ("__init__", "__main__", "__app__"):
            name = pkg

    return structlog.get_logger(name, *args, **initial_values)


__all__ = ["get_logger"]



from harp.utils.testing.config import BaseConfigurableTest
from harp_apps.proxy.settings.remote.probe import RemoteProbe, RemoteProbeSettings


class TestRemoteProbeSettings(BaseConfigurableTest):
    type = RemoteProbeSettings
    initial = {}
    expected = {}
    expected_verbose = {
        "headers": {},
        "interval": 10.0,
        "method": "GET",
        "path": "/",
        "timeout": 10.0,
        "verify": True,
    }


class TestRemoteProbeStateful(BaseConfigurableTest):
    type = RemoteProbe
    initial = {"settings": TestRemoteProbeSettings.initial}
    expected = {"settings": TestRemoteProbeSettings.expected_verbose}
    expected_verbose = {"settings": TestRemoteProbeSettings.expected_verbose}



"""
Telemetry application.

Sends anonymous usage statistics on startup.

todo:
- send ping every 24 hours
- send ping on exit ?
- activity type
- license key
- send platform to allow to understand target env types (linux, windows, macos, what python, etc.)
- detect if within container ?
- detect if within kubernetes ?
- detect git/docker/pip/brew install ?

"""

from harp import get_logger
from harp.config import Application
from harp.config.events import OnBindEvent, OnBoundEvent
from harp_apps.telemetry.manager import TelemetryManager

logger = get_logger(__name__)


async def on_bind(event: OnBindEvent):
    event.container.add_singleton(TelemetryManager)


async def on_bound(event: OnBoundEvent):
    event.provider.get(TelemetryManager)


application = Application(
    on_bind=on_bind,
    on_bound=on_bound,
)



from harp.config import Application
from harp.config.asdict import asdict
from harp_apps.storage.settings import StorageSettings


def test_empty_settings():
    settings = StorageSettings()

    assert asdict(settings, verbose=True) == {
        "migrate": True,
        "url": "sqlite+aiosqlite:///:memory:?cache=shared",
        "blobs": {"type": "sql"},
        "redis": None,
    }

    assert asdict(settings) == {}


def test_secure():
    settings = StorageSettings(url="postgresql://user:password@localhost:5432/db")

    assert asdict(settings, verbose=True, mode="python") == {
        "migrate": True,
        "url": "postgresql+asyncpg://user:***@localhost:5432/db",
        "blobs": {"type": "sql"},
        "redis": None,
    }

    assert asdict(settings, mode="python") == {
        "url": "postgresql+asyncpg://user:***@localhost:5432/db",
    }

    assert asdict(settings, verbose=True, secure=False) == {
        "migrate": True,
        "url": "postgresql+asyncpg://user:password@localhost:5432/db",
        "blobs": {"type": "sql"},
        "redis": None,
    }

    assert asdict(settings, secure=False) == {
        "url": "postgresql+asyncpg://user:password@localhost:5432/db",
    }


def test_override_blob_storage_type():
    settings = StorageSettings.from_kwargs(blobs={"type": "redis"})
    assert asdict(settings, verbose=True) == {
        "migrate": True,
        "url": "sqlite+aiosqlite:///:memory:?cache=shared",
        "blobs": {"type": "redis"},
        "redis": None,
    }

    assert asdict(settings) == {
        "blobs": {"type": "redis"},
    }


def test_override_redis_url():
    settings = StorageSettings.from_kwargs(blobs={"type": "redis"}, redis={"url": "redis://example.com:1234/42"})
    assert asdict(settings, verbose=True) == {
        "blobs": {"type": "redis"},
        "migrate": True,
        "redis": {"url": "redis://example.com:1234/42"},
        "url": "sqlite+aiosqlite:///:memory:?cache=shared",
    }
    assert asdict(settings) == {
        "blobs": {"type": "redis"},
        "redis": {"url": "redis://example.com:1234/42"},
    }


def test_settings_normalization_does_not_hide_password():
    app = Application(settings_type=StorageSettings)
    settings = app.normalize({"url": "postgresql://user:password@localhost:5432/db"})
    assert settings["url"] == "postgresql+asyncpg://user:password@localhost:5432/db"



import json


class BytesEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, bytes):
            return o.decode("ascii")
        else:
            return super().default(o)



from unittest import mock

import pytest

from harp_apps.notifications.senders import GoogleChatNotificationSender
from harp_apps.notifications.senders.google_chat import AsyncClient


@pytest.mark.parametrize(
    "public_url, expected_link",
    [
        ("https://example.com", True),
        (None, False),
    ],
)
def test_message_does_not_contain_link_if_no_public_url(public_url, expected_link):
    notification_sender = GoogleChatNotificationSender("https://slack.com/webhook")

    message = notification_sender._format_error_message(
        "GET", "https://example.com", 500, "Server Error", "123", public_url
    )

    sections = message["cards"][0]["sections"]

    def widget_contains_buttons(section):
        widgets = section.get("widgets")
        return any("buttons" in widget for widget in widgets)

    # check that block does not contain any "action" typed elements
    link_button = any(widget_contains_buttons(section) for section in sections)

    assert link_button == expected_link


async def test_send_notification():
    with mock.patch.object(AsyncClient, "post", new_callable=mock.AsyncMock) as mock_post:
        webhook_url = "http://example.com/webhook"
        public_url = "http://example.com"
        sender = GoogleChatNotificationSender(webhook_url, public_url)

        mock_post.return_value.status_code = 200

        method = "POST"
        url = "http://example.com/api"
        status_code = 500
        message = "Internal Server Error"
        transaction_id = "12345"

        await sender.send_notification(method, url, status_code, message, transaction_id)
        mock_post.assert_called_once()
        assert mock_post.call_args[0][0] == webhook_url



import ast
from typing import Callable


def _get_normalized_sources_from_ast(code_ast: ast.AST, /):
    """

    :param code_ast: abstract syntax tree to convert to source code
    :return: noramlized source code string
    """
    statements = []
    for node in code_ast.body:
        statements.append(ast.unparse(node))
    return ("\n".join(statements)).strip()


class Script(Callable):
    """
    A script object represents a small python script that can be executed in a controlled environment. For example, it
    will hold your scripts defined in rules.
    """

    def __init__(self, source: ast.AST | str = None, /, *, filename=None):
        self._filename = filename or f"<{type(self).__name__.lower()}:{id(self)}>"
        self._source = None

        if source is None:
            raise ValueError("Code or AST must be provided.")

        if isinstance(source, str):
            source = ast.parse(source)

        self._code = compile(source, filename=self._filename, mode="exec")
        self._source = _get_normalized_sources_from_ast(source)

    @property
    def source(self):
        return self._source

    @property
    def filename(self):
        return self._filename

    @classmethod
    def from_file(cls, filename: str):
        """Constructor from a file content, by filename."""
        with open(filename) as f:
            return cls(f.read(), filename=filename)

    def __call__(self, context: dict):
        exec(self._code, None, context)

    def __eq__(self, other):
        return isinstance(other, Script) and self._code == other._code

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        _type = type(self).__name__
        _source = self.source
        return f"{_type}({repr(_source)}, filename={repr(self.filename)})"


class ExecutableObject(Callable):
    def __init__(self, target: Callable):
        self._target = target

    def __call__(self, context: dict):
        return self._target(context)



import pytest
from multidict import MultiDict

from harp.utils.collections import MultiChainMap


class TestMultiChainMap:
    def create_from_two_multidicts(self):
        child1 = MultiDict((("a", 1), ("b", 2), ("a", 3)))
        child2 = MultiDict((("b", 4), ("c", 5)))
        return MultiChainMap(child1, child2)

    def test_get_one_dict(self):
        c = MultiChainMap({"a": 1, "b": 2})
        assert c["a"] == 1
        assert c["b"] == 2
        with pytest.raises(KeyError):
            _ = c["c"]

    def test_get_two_dicts(self):
        c = MultiChainMap({"a": 1, "b": 2}, {"b": 3, "c": 4})
        assert c["a"] == 1
        assert c["b"] == 2
        assert c["c"] == 4
        with pytest.raises(KeyError):
            _ = c["d"]

    def test_empty(self):
        c = MultiChainMap()
        with pytest.raises(KeyError):
            _ = c["a"]

    def test_get_from_one_multi_dict(self):
        child = MultiDict((("a", 1), ("b", 2), ("a", 3)))
        wrapper = MultiChainMap(child)
        assert wrapper["a"] == 1
        assert wrapper["b"] == 2
        with pytest.raises(KeyError):
            _ = wrapper["c"]

    def test_get_from_two_multi_dict(self):
        c = self.create_from_two_multidicts()
        assert c["a"] == 1
        assert c["b"] == 2
        assert c["c"] == 5
        with pytest.raises(KeyError):
            _ = c["d"]

        assert dict(c) == {"a": 1, "b": 2, "c": 5}

    def test_items(self):
        c = self.create_from_two_multidicts()
        assert list(c.items()) == [("b", 2), ("c", 5), ("a", 1), ("a", 3)]

    def test_iter(self):
        c = self.create_from_two_multidicts()

        ci = iter(c)
        assert next(ci) == "b"
        assert next(ci) == "c"
        assert next(ci) == "a"
        assert next(ci) == "a"
        with pytest.raises(StopIteration):
            next(ci)

    def test_len(self):
        assert len(self.create_from_two_multidicts()) == 4

    def test_contains(self):
        c = self.create_from_two_multidicts()
        assert "a" in c
        assert "b" in c
        assert "c" in c
        assert "d" not in c

    def test_notset(self):
        c = self.create_from_two_multidicts()

        assert "b" in c
        del c["b"]
        assert "b" not in c

        # "b" source was not touched
        assert "b" in c.maps[1]



from harp.services.models import _resolve
from harp.services.references import LazySettingReference


def test_basic_cfgref():
    ref = LazySettingReference("foo.bar")
    assert ref.resolve({"foo": {"bar": "baz"}}) == "baz"


def test_equality_cfgref():
    ref = LazySettingReference("foo.bar == 'baz'")
    assert ref.resolve({"foo": {"bar": "baz"}}) is True
    assert ref.resolve({"foo": {"bar": "boo"}}) is None


def test_resolve_dict_with_references():
    arguments = {
        "foobar": LazySettingReference("foo.bar"),
    }
    assert _resolve(arguments, {"foo": {"bar": "baz"}}) == {"foobar": "baz"}


def test_resolve_dict_with_deep_references():
    arguments = {
        "foo": {"bar": LazySettingReference("foo.bar")},
    }
    assert _resolve(arguments, {"foo": {"bar": "baz"}}) == {"foo": {"bar": "baz"}}



from unittest import mock

import pytest

from harp_apps.notifications.senders import SlackNotificationSender
from harp_apps.notifications.senders.slack import AsyncWebhookClient


@pytest.mark.parametrize(
    "public_url, expected_link",
    [
        ("https://example.com", True),
        (None, False),
    ],
)
def test_message_does_not_contain_link_if_no_public_url(public_url, expected_link):
    notification_sender = SlackNotificationSender("https://slack.com/webhook")

    message = notification_sender._format_error_message(
        "GET", "https://example.com", 500, "Server Error", "123", public_url
    )

    blocks = message["blocks"]

    # check that block does not contain any "action" typed elements
    link_button = any(v == "actions" for block in blocks for k, v in block.items())

    assert link_button == expected_link


async def test_send_notification():
    with mock.patch.object(AsyncWebhookClient, "send", new_callable=mock.AsyncMock) as mock_send:
        webhook_url = "http://example.com/webhook"
        public_url = "http://example.com"
        sender = SlackNotificationSender(webhook_url, public_url)

        mock_send.return_value.status_code = 200
        mock_send.return_value.body = "ok"

        method = "POST"
        url = "http://example.com/api"
        status_code = 500
        message = "Internal Server Error"
        transaction_id = "12345"

        await sender.send_notification(method, url, status_code, message, transaction_id)
        mock_send.assert_called_once()
        assert mock_send.call_args[1]["text"] == "ERROR NOTIFICATION"



from typing import Any


def import_string(type: str) -> Any:
    _path, _attr = type.rsplit(".", 1)
    concrete_type = getattr(__import__(_path, fromlist=[_attr]), _attr)
    return concrete_type


def get_module_name(_type):
    return _type.__module__


def get_qualified_name(_type):
    try:
        return _type.__qualname__
    except AttributeError:
        return _type.__name__


def get_full_qualified_name(_type):
    return f"{get_module_name(_type)}.{get_qualified_name(_type)}"



from typing import Protocol


class MessageSerializer(Protocol):
    @property
    def summary(self) -> str:
        return ...

    @property
    def headers(self) -> str:
        return ...

    @property
    def body(self) -> bytes:
        return ...



from config.common import merge_values


def test_merge_values_first_level():
    a = {"rules": {"a": 1, "b": 2}}
    b = {"rules": {"b": 3, "c": 4}}

    merge_values(a, b)
    assert a == {"rules": {"a": 1, "b": 3, "c": 4}}


def test_merge_values_deep_list():
    a = {"rules": {"*": {"*": ["foo"]}}}
    b = {"rules": {"*": {"*": "bar"}}}

    merge_values(a, b)
    assert a == {"rules": {"*": {"*": "bar"}}}



import dataclasses
from datetime import datetime
from typing import List

from .base import Entity
from .messages import Message


@dataclasses.dataclass(kw_only=True)
class Transaction(Entity):
    id: str = None
    """Unique identifier for this transaction."""

    type: str  # enum http websocket lifecycle ...
    """Type of ASGI transaction: http, websocket, lifecycle, ..."""

    endpoint: str = None
    """Endpoint name, if any (this describes which proxy controller handled the request)."""

    started_at: datetime
    """Timestamp of the transaction start."""

    finished_at: datetime = None
    """Timestamp of the transaction end (if it has ended), or None."""

    # Computed fields

    elapsed: float = None
    """Elapsed time in seconds, if the transaction has ended."""

    tpdex: int = None
    """TPDEX score, if the transaction has ended."""

    # Relations

    messages: List[Message] = None
    tags: dict = dataclasses.field(default_factory=dict)

    # Extra attributes
    extras: dict = dataclasses.field(default_factory=dict)

    # Markers, used for transaction management (e.g. skip storage on high pressure)
    markers: set = dataclasses.field(default_factory=set)

    def as_storable_dict(self, /, *, with_messages=False, with_tags=False):
        """Create a dict that can be passed to a storage creation method."""
        data = self._asdict()

        if not with_messages:
            data.pop("messages", None)

        if not with_tags:
            data.pop("tags", None)

        extras = data.pop("extras", {})

        data["x_cached"] = extras.get("cached")
        data["x_method"] = extras.get("method")
        data["x_no_cache"] = bool(extras.get("no_cache"))
        data["x_status_class"] = extras.get("status_class")

        return data

    def _asdict(self, /, *, secure=True):
        result = dataclasses.asdict(self)
        del result["markers"]
        return result



import asyncio
from typing import List, Optional

from whistle import IAsyncEventDispatcher

from harp.asgi.events import EVENT_CORE_RESPONSE, ResponseEvent
from harp_apps.notifications.senders.google_chat import GoogleChatNotificationSender
from harp_apps.notifications.senders.slack import SlackNotificationSender
from harp_apps.notifications.settings import NotificationsSettings
from harp_apps.notifications.typing import NotificationSender

AVAILABLE_WEBHOOK_URLS = ["slack_webhook_url", "google_chat_webhook_url"]


class NotificationSubscriber:
    def __init__(self, settings: NotificationsSettings, public_url: Optional[str] = None):
        self.senders: List[NotificationSender] = []
        for name in AVAILABLE_WEBHOOK_URLS:
            if name == "google_chat_webhook_url":
                try:
                    webhook_url = getattr(settings, name)
                    (
                        self.senders.append(GoogleChatNotificationSender(webhook_url, public_url))
                        if webhook_url
                        else None
                    )
                except AttributeError:
                    pass
            elif name == "slack_webhook_url":
                try:
                    webhook_url = getattr(settings, name)
                    (self.senders.append(SlackNotificationSender(webhook_url, public_url)) if webhook_url else None)
                except AttributeError:
                    pass

    def subscribe(self, dispatcher: IAsyncEventDispatcher):
        dispatcher.add_listener(EVENT_CORE_RESPONSE, self.on_response_send_error_notifications)

    def unsubscribe(self, dispatcher: IAsyncEventDispatcher):
        dispatcher.remove_listener(EVENT_CORE_RESPONSE, self.on_response_send_error_notifications)

    async def on_response_send_error_notifications(self, event: ResponseEvent):
        if 500 <= event.response.status < 600:
            transaction = event.request.extensions.get("transaction")
            await self.send_notification(
                method=event.request.extensions.get("remote_method"),
                url=event.request.extensions.get("remote_url"),
                status_code=event.response.status,
                message=event.response.reason_phrase,
                transaction_id=transaction.id if transaction else None,
            )

    async def send_notification(
        self,
        method: Optional[str],
        url: Optional[str],
        status_code: int,
        message: str,
        transaction_id: Optional[str],
    ) -> None:
        async with asyncio.TaskGroup() as tg:
            for sender in self.senders:
                tg.create_task(sender.send_notification(method, url, status_code, message, transaction_id))



from harp.http import HttpRequest
from harp_apps.proxy.events import ProxyFilterEvent
from harp_apps.rules import examples
from harp_apps.rules.settings import RulesSettings


class TestTeapotExample:
    source = examples.load("teapot.yml").get("rules")

    def create_request(self, **kwargs):
        return HttpRequest(**kwargs)

    def create_proxy_filter_event(self, event_name, /, *, endpoint, **kwargs):
        event = ProxyFilterEvent(endpoint, **kwargs)
        event.name = event_name
        return event

    def get_rules(self):
        settings = RulesSettings(**self.source)
        return settings.ruleset

    def test_on_proxy_filter_request(self):
        rules = self.get_rules()
        event = self.create_proxy_filter_event("proxy.filter.request", endpoint="acme1", request=self.create_request())

        scripts = list(rules.match(*event.criteria))
        assert len(scripts) == 1

        event.execute_script(scripts[0])
        assert event.response.status == 418



import copy

import pytest

from harp.config.asdict import asdict
from harp.utils.testing.applications import BaseTestForApplications


class TestStorageApplication(BaseTestForApplications):
    name = "harp_apps.storage"
    config_key = "storage"

    expected_defaults = {
        "blobs": {"type": "sql"},
        "migrate": True,
        "redis": None,
        "url": "sqlite+aiosqlite:///:memory:?cache=shared",
    }

    @pytest.mark.parametrize(
        ["settings"],
        [
            [{"url": "sqlite+aiosqlite:///harp.db"}],
            [{"migrate": True}],
            [{"migrate": False}],
            [
                {
                    "blobs": {"type": "redis"},
                    "redis": {"url": "redis://localhost:6379/1"},
                }
            ],
        ],
    )
    def test_defaults_fills_missing_values_for_sqlalchemy_type(self, settings: dict):
        application_settings = self.application.settings_type(**copy.deepcopy(settings))
        assert asdict(application_settings, verbose=True) == {
            **self.expected_defaults,
            **settings,
        }



from functools import cached_property
from typing import Annotated, Any, Literal

from pydantic import BeforeValidator, Field, model_validator

from harp.config import Configurable
from harp.utils.config import yaml
from harp_apps.dashboard.constants import TAlgorithm


def _replace_plain_algorithm_for_bc(value):
    if value == "plain":
        return "plaintext"
    return value


def _load(value):
    if isinstance(value, dict) and "fromFile" in value:
        with open(value["fromFile"]) as f:
            return yaml.safe_load(f)
    return value


class User(Configurable):
    password: str

    @model_validator(mode="before")
    @classmethod
    def _replace_strings_by_password_dict(cls, data: Any) -> Any:
        if isinstance(data, str):
            return {"password": data}
        return data


class BasicAuthSettings(Configurable):
    type: Literal["basic"] = Field(
        None,
        description="Authentication type. Only «basic» is supported for now.",
    )

    algorithm: Annotated[
        TAlgorithm,
        BeforeValidator(_replace_plain_algorithm_for_bc),
        Field(
            "pbkdf2_sha256",
            description="Hashing algorithm used for passwords.",
        ),
    ]

    users: Annotated[
        dict[str, User],
        BeforeValidator(_load),
        Field(
            default_factory=dict,
            description="Users list.",
        ),
    ]

    @cached_property
    def algorithm_impl(self):
        impl = __import__("passlib.hash", fromlist=[self.algorithm])
        return getattr(impl, self.algorithm).verify

    def check(self, username, password):
        user = self.users.get(username)

        if not user:
            return False

        if not self.algorithm_impl(password, user.password):
            return False

        return username

    @model_validator(mode="before")
    @classmethod
    def model_always_shown_defaults(cls, data):
        """This validator overcomes the behaviour of model_dump(exclude_defaults=True, exclude_unset=True) and forces
        some defaults to be dumped everytime."""
        data.setdefault("type", "basic")
        return data



import dataclasses
from typing import Union

from harp.typing import Maybe


@dataclasses.dataclass(kw_only=True, frozen=True)
class HttpMethod:
    request_body: bool
    response_body: Union[bool, Maybe]
    safe: bool
    idempotent: bool
    cacheable: Union[bool, Maybe]
    allowed_in_forms: bool
    description: str = None
    link: str = None


HTTP_METHODS = {
    "GET": HttpMethod(
        request_body=False,
        response_body=True,
        safe=True,
        idempotent=True,
        cacheable=True,
        allowed_in_forms=True,
        description=(
            "The GET method requests a representation of the specified resource. Requests using GET should only "
            "retrieve data."
        ),
        link="https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/GET",
    ),
    "HEAD": HttpMethod(
        request_body=False,
        response_body=False,
        safe=True,
        idempotent=True,
        cacheable=True,
        allowed_in_forms=False,
        description="The HEAD method asks for a response identical to a GET request, but without the response body.",
        link="https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/HEAD",
    ),
    "POST": HttpMethod(
        request_body=True,
        response_body=True,
        safe=False,
        idempotent=False,
        cacheable=Maybe,
        allowed_in_forms=True,
        description=(
            "The POST method submits an entity to the specified resource, often causing a change in state or side "
            "effects on the server."
        ),
        link="https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/POST",
    ),
    "PUT": HttpMethod(
        request_body=True,
        response_body=Maybe,
        safe=False,
        idempotent=True,
        cacheable=False,
        allowed_in_forms=False,
        description=(
            "The PUT method replaces all current representations of the target resource with the request payload."
        ),
        link="https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/PUT",
    ),
    "DELETE": HttpMethod(
        request_body=False,
        response_body=Maybe,
        safe=False,
        idempotent=True,
        cacheable=False,
        allowed_in_forms=False,
        description="The DELETE method deletes the specified resource.",
        link="https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/DELETE",
    ),
    "CONNECT": HttpMethod(
        request_body=False,
        response_body=False,
        safe=False,
        idempotent=False,
        cacheable=False,
        allowed_in_forms=False,
        description="The CONNECT method establishes a tunnel to the server identified by the target resource.",
        link="https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/CONNECT",
    ),
    "OPTIONS": HttpMethod(
        request_body=False,
        response_body=True,
        safe=True,
        idempotent=True,
        cacheable=False,
        allowed_in_forms=False,
        description="The OPTIONS method describes the communication options for the target resource.",
        link="https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/OPTIONS",
    ),
    "TRACE": HttpMethod(
        request_body=False,
        response_body=False,
        safe=True,
        idempotent=True,
        cacheable=False,
        allowed_in_forms=False,
        description="The TRACE method performs a message loop-back test along the path to the target resource.",
        link="https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/TRACE",
    ),
    "PATCH": HttpMethod(
        request_body=True,
        response_body=Maybe,
        safe=False,
        idempotent=False,
        cacheable=Maybe,
        allowed_in_forms=False,
        description="The PATCH method applies partial modifications to a resource.",
        link="https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/PATCH",
    ),
}



from unittest.mock import Mock

from httpx import AsyncClient

from .._base import BaseRulesFlowTest


class TestHttpClientRulesFlow(BaseRulesFlowTest):
    applications = ["http_client", "rules"]

    async def test_http_client_flow(self, httpbin):
        mock = Mock()
        system = await self.create_system({}, mock=mock)

        # make a request
        http_client = system.provider.get(AsyncClient)
        response = await http_client.get(httpbin)
        assert response.status_code == 200

        assert mock.call_count == 2

        assert mock.call_args_list[0].args[0]["rule"] == "on_remote_request"
        assert mock.call_args_list[0].args[0]["response"] is None

        assert mock.call_args_list[1].args[0]["rule"] == "on_remote_response"
        assert mock.call_args_list[1].args[0]["response"].status_code == 200



from typing import TYPE_CHECKING, List

from sqlalchemy import Integer, String, select
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import Mapped, mapped_column, relationship

from . import Base
from .base import Repository

if TYPE_CHECKING:
    from .flags import UserFlag


class User(Base):
    __tablename__ = "users"

    id = mapped_column(Integer(), primary_key=True, unique=True, autoincrement=True)
    username = mapped_column(String(32), unique=True)

    flags: Mapped[List["UserFlag"]] = relationship(
        back_populates="user",
        cascade="all, delete",
        passive_deletes=True,
    )


class UsersRepository(Repository):
    Type = User

    async def find_one_by_username(self, username: str) -> User:
        async with self.session_factory() as session:
            try:
                return (await session.execute(select(self.Type).where(User.username == username))).unique().scalar_one()
            except NoResultFound:
                return (
                    (await session.execute(select(self.Type).where(User.username == "anonymous"))).unique().scalar_one()
                )



import os

from yaml import load

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

import harp
from harp_apps.rules.settings import RulesSettings

with open(os.path.join(harp.ROOT_DIR, "docs/apps/rules/examples/rules.yml")) as f:
    example_rules = load(f, Loader=Loader).get("rules")


def test_idempotence():
    settings = RulesSettings(**example_rules)
    normalized_sources = settings._asdict()
    assert normalized_sources == {
        "*": {
            "*": {
                "on_request": "request.headers['X-Forwarded-For'] = 'Joe'",
            }
        },
        "httpbin-*": {
            "GET /*": {
                "on_remote_response": "response.headers['Cache-Control'] = 'max-age=3600'",
            }
        },
    }

    settings = RulesSettings(**normalized_sources)
    assert normalized_sources == settings._asdict()


def test_apply():
    settings = RulesSettings(
        **{
            "*": {
                "*": {
                    "*": ["foo = 42"],
                }
            }
        }
    )

    context = {}
    for script in settings.ruleset.match("api1", "GET /foo", "on_request"):
        script(context)
    assert context == {"foo": 42}



from typing import Optional
from urllib.parse import parse_qsl

from asgiref.typing import ASGIReceiveCallable, HTTPScope
from httpx import AsyncByteStream
from multidict import CIMultiDict, MultiDict

from harp.asgi.bridge.streams import AsyncStreamFromAsgiReceive
from harp.http import HttpRequestBridge

_default_host = (None, None)


class HttpRequestAsgiBridge(HttpRequestBridge):
    """Actually implements the getters required by HttpRequest using the asgi scope and receive callable. It is still
    an early implementation and will need to support streaming requests in the future.
    """

    def __init__(self, scope: HTTPScope, receive: ASGIReceiveCallable):
        """
        :param scope: see https://asgi.readthedocs.io/en/latest/specs/www.html#http-connection-scope
        :param receive:
        """
        self.asgi_scope = scope
        self.asgi_receive = receive

        self._closed = False

    def get_server_ipaddr(self) -> Optional[str]:
        """Get the server IP address from asgi scope."""
        try:
            return (self.asgi_scope.get("server", _default_host) or _default_host)[0]
        except IndexError:
            return None

    def get_server_port(self) -> Optional[int]:
        """Get the server port from asgi scope."""
        try:
            return (self.asgi_scope.get("server", _default_host) or _default_host)[1]
        except IndexError:
            return None

    def get_method(self) -> str:
        """Get the HTTP method from asgi scope."""
        return self.asgi_scope.get("method", None)

    def get_path(self) -> str:
        """Get the HTTP path from asgi scope (/foo/bar), without the query string part."""
        if "raw_path" in self.asgi_scope:
            return self.asgi_scope["raw_path"].decode("utf-8")
        return self.asgi_scope.get("path", "/")

    def get_query(self) -> MultiDict:
        """Get the query string from asgi scope, as a multidict."""
        return MultiDict(
            parse_qsl(
                self.asgi_scope.get("query_string", b"").decode("utf-8"),
                keep_blank_values=True,
            )
        )

    def get_headers(self) -> CIMultiDict:
        """Get the headers from asgi scope, as a case-insensitive multidict."""
        headers = CIMultiDict()
        for name, value in self.asgi_scope.get("headers", []):
            headers.add(name.decode("utf-8"), value.decode("utf-8"))
        return headers

    def get_stream(self) -> AsyncByteStream:
        """Get the request body stream from asgi receive callable."""
        return AsyncStreamFromAsgiReceive(self.asgi_receive)



from datetime import datetime, timedelta

from harp_apps.storage.types import IStorage

from .base import FacetWithStorage


class TransactionEndpointFacet(FacetWithStorage):
    name = "endpoint"

    def __init__(self, *, storage: IStorage):
        super().__init__(storage=storage)
        self.choices = set()
        self._refreshed_at = None

    async def refresh(self):
        if not self._refreshed_at or (self._refreshed_at < datetime.now() - timedelta(seconds=25)):
            self._refreshed_at = datetime.now()
            meta = await self.storage.get_facet_meta(self.name)
            self.choices = set(meta.keys())
            self.meta = {endpoint: {"count": count} for endpoint, count in meta.items()}
